/* $NetBSD$ */

/*
 * File "udf.c" is part of the UDFclient toolkit.
 * File $Id: udf.c,v 1.74 2003/09/14 14:19:00 imago Exp $ $Name: release-ro1 $
 *
 * Copyright (c) 2003 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udfclient/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <limits.h>

#include "scsi_sense.h"
#include "scsilib.h"


/* for locals */
#include "udf.h"
#include "udf_bswap.h"


/* for scsilib */
const char *dvname="UDF device";


#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))


//#define DEBUG(a) { a; };
#define DEBUG(a) if (0) { a; };

#if 1
	extern void udf_dump_descriptor(union dscrptr *dscrpt);
	extern void udf_dump_vat_table(struct udf_part_mapping *udf_part_mapping);
	extern void udf_dump_disc_anchors(struct udf_discinfo *disc);
	extern void udf_dump_alive_sets(void);
#else
#define udf_dump_descriptor(a)
#define udf_dump_vat_table(a)
#define udf_dump_disc_anchors(a)
#define udf_dump_alive_sets()
#endif


/* globals */
struct udf_discinfo  *udf_discs_list     = NULL;
struct udf_volumeset *udf_volumeset_list = NULL;
struct hash_table    *udf_mountables     = NULL;
int udf_verbose;

extern int errno;


/******************************************************************************************
 *
 * Tables and helper functions section
 *
 ******************************************************************************************/

int read_cd_hex2(int val) {
	int nl, nh;

	nl = val & 15;
	nh = val >> 4;
	if (nl >= 'A') nl -= 'A' + 10;
	if (nh >= 'A') nh -= 'A' + 10;

	return (nh*16) + nl;
}


int read_cd_bcd(int val) {
	int nl, nh;

	nl = (val & 15) - '0';
	nh = (val >> 4) - '0';
	if ((nl < 0 || nl > 9) || (nh < 0 || nh > 9)) return val;

	return nh*10 + nl;
}


int cd_msf2lba(int m, int s, int f) {
	return 4500*m + 75*s + f - 150;
}


/******************************************************************************************
 *
 * Disc access abstraction section
 *
 ******************************************************************************************/

#define SCSI_CMD_LEN  12
typedef uint8_t scsicmd[SCSI_CMD_LEN];


int udf_discinfo_get_type(struct udf_discinfo *disc) {
	return (disc->devdrv_class << 8) | (disc->recordable << 4) | (disc->rewritable);
}


int scsi_call(int flags, int fileh, void *cmd, int cmdlen, void *data, int datalen) {
#ifdef SCSI
	return scsi_command(fileh, cmd, cmdlen, data, datalen, 30000, flags);
#else
	return ENODEV;
#endif
}


#ifndef SCSI
int scsi_check_for_scsi(int fhandle) {
	return ENODEV;
}
#endif


/*
 * This routine 'get disc type' is a bit 'magic' in that it tries to figure out what
 * kind of device/disc we are dealing with without going into obscure functions to
 * let it be sorted out for us; ONLY fills in devdrv_class and MMC profile.
 */

int udf_discinfo_get_disc_type(struct udf_discinfo *disc) {
	struct stat stat;
	scsicmd	    cmd;
	uint8_t	    buf[40];

	/* assume generic CD-ROM with no known MMC profile */
	disc->devdrv_class = UDF_DEVDRV_CLASS_CD;
	disc->mmc_profile  = 0;

	/* check if its a regular file */
	fstat(disc->fhandle, &stat);
	if (S_ISREG(stat.st_mode)) {
		UDF_VERBOSE(printf("UDF device %s is a regular file\n", disc->dev_name));
		disc->devdrv_class   = UDF_DEVDRV_CLASS_FILE;
		return 0;
	};

	/* check if its a ATAPI/SCSI device */
	errno = scsi_check_for_scsi(disc->fhandle);
	if (errno) {
		/* obviously no ATAPI/SCSI device -> has to be IDE disc or other but no file either */
		disc->devdrv_class   = UDF_DEVDRV_CLASS_DISC;
		UDF_VERBOSE(printf("Got error executing SCSI command, assuming IDE disc\n"));
		return 0;
	};

	/* get MMC profile */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x46;				/* Get configuration */
	cmd[ 8] = 32;				/* just a small buffer size */
	cmd[ 9] = 0;				/* control */
	errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, buf, 32);
	if (!errno) {
		disc->mmc_profile = buf[7] | (buf[6] << 8);
	} else {
		disc->mmc_profile = 0;	/* mark unknown MMC profile */
	};
	UDF_VERBOSE_MAX(printf("Device has MMC profile 0x%02x\n", disc->mmc_profile));

	/* try to be smart */
	switch (disc->mmc_profile) {
		case 0x01 : case 0x02 : case 0x03 : case 0x04 : case 0x05 :
			/* SCSI discs class; treat like normal discs */
			disc->devdrv_class   = UDF_DEVDRV_CLASS_DISC;
			UDF_VERBOSE(printf("SCSI disc detected; treat like normal disc device\n"));
			return 0;
		case 0x00 : case 0x08 : case 0x09 : case 0x0a :
			/* different types of CD-ROM devices like CD-R/RW etc. */
			disc->devdrv_class = UDF_DEVDRV_CLASS_CD;
			break;
		case 0x10 : case 0x11 : case 0x12 : case 0x13 : case 0x14 : case 0x1a : case 0x1b : case 0x20 :
			/* various DVD+/- R/RW devices */
			disc->devdrv_class = UDF_DEVDRV_CLASS_DVD;
			break;
		default :
			break;
	};

	/* We are now sure its either a CD or a DVD ; but we might still be on a legacy CD-ROM drive */
	return 0;
}


/* 10000 is arbitrary */
int udf_get_disc_info(struct udf_discinfo *disc) {
	scsicmd		 cmd;
	struct stat	 stat;
	uint8_t		 res[10000];
	uint32_t	 res_len = 10000;
	uint32_t	 first_track, last_track;
	uint32_t	 first_session, last_session;
	uint32_t	 first_track_last_session, last_track_last_session;
	uint32_t	 nwa_valid, next_writable_addr;
	uint32_t	 track, session;
	uint32_t	 cntrl, addr, tno, point, min, sec, frame, pmin, psec, pframe;
	uint32_t	 data_length, pos;
	off_t		 track_start, track_end, track_size, disc_size;

	errno = udf_discinfo_get_disc_type(disc);
	if (errno) {
		perror("Error during classification of disc; skipping disc\n");
		return errno;
	};

	if (disc->devdrv_class == UDF_DEVDRV_CLASS_FILE) {
		fstat(disc->fhandle, &stat);
		disc->recordable         = 1;					/* assuming rw access	*/
		disc->rewritable         = 1;
		disc->sector_size	 = UDF_SECTOR_SIZE;			/* take a value; not applicable	really */
		disc->disc_state         = DISC_STATE_NOT_SERIAL;
		disc->last_session_state = SESSION_STATE_INCOMPLETE;
		disc->num_sessions       = 1;
		disc->session_start [0]  = 0;
		disc->session_end   [0]  =  stat.st_size    / UDF_SECTOR_SIZE;	/* sectors right? */
		disc->next_writable [0]  = (stat.st_size    / UDF_SECTOR_SIZE) + 1;
		disc->packet_size   [0]  =  stat.st_blksize / UDF_SECTOR_SIZE;
		return 0;
	};

	if (disc->devdrv_class == UDF_DEVDRV_CLASS_DISC) {
		disc->recordable         = 1;
		disc->rewritable         = 1;
		disc->sector_size	 = DISC_SECTOR_LENGTH;	/* XXX normal on Unix systems; assumption!!! XXX */
		disc->disc_state         = DISC_STATE_NOT_SERIAL;
		disc->last_session_state = SESSION_STATE_COMPLETE;
		disc->num_sessions       = 1;
		disc->session_start [0]  = 0;
		disc->session_end   [0]  = 0;	/* XXX how to determine? ; no VAT for now XXX */
		fprintf(stderr, "UDF: warning... reading/writing on 'disc' device\n");
		fprintf(stderr, "UDF: warning... assuming sector_size; of %d bytes\n", DISC_SECTOR_LENGTH);
		return 0;
	}

	/* classes CD and DVD remain; we can be on a DVD/CD recordable or on a legacy CD-ROM */

	/* Set CD/DVD speed on 'optimal' for it doesnt seem to do it automatically */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0xBB;				/* Set CD speed */
	cmd[ 1] = 0;				/* select CLV recording */
	cmd[ 2] = 0xff;
	cmd[ 3] = 0xff;				/* max read performance speed */
	cmd[ 4] = 0xff;
	cmd[ 5] = 0xff;				/* max write performance speed; applic? */
	cmd[11] = 0;				/* control */
	scsi_call(SCSI_WRITECMD, disc->fhandle, cmd, 12, res, res_len);

	/* Get track information */
	bzero(res, res_len);
	bzero(cmd, SCSI_CMD_LEN);

	cmd[0] = 0x51;				/* Read disc information */
	cmd[7] = res_len >> 8;			/* MSB allocation length */
	cmd[8] = res_len & 0xff;		/* LSB allocation length */
	cmd[9] = 0;				/* control */
	errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, res, res_len);

	if (!errno) {
		/* we are a MMC3+ device! */
		disc->disc_state		 =  res[2]       &  3;		/* just 'happends' to be the same as we use */
		disc->last_session_state	 = (res[2] >> 2) &  3;		/* ditto */
		disc->recordable		 =  1;				/* XXX this might be proven wrong one day XXX */
		disc->rewritable		 =  res[2]       & 16;		/* erasable */
		disc->sector_size		 =  CD_SECTOR_LENGTH;
		disc->num_sessions		 =  res[4] | (res[ 9] << 8);
		first_track			 =  res[3];
		first_track_last_session	 =  res[5] | (res[10] << 8);	/* to build up the last 'session' */
		last_track_last_session  	 =  res[6] | (res[11] << 8);

		/* Initialise the sessions to be taken as overspanning tracks */
		for(session = 0; session < disc->num_sessions; session++) {
			disc->session_start[session] = INT_MAX;
			disc->session_end  [session] = 0;
		}

		for (track = first_track; track <= last_track_last_session; track++) {
			bzero(cmd, SCSI_CMD_LEN);
			bzero(res, res_len);
			cmd[0] = 0x52;				/* Read track informareson */
			cmd[1] = 1;				/* indexed on track       */
			cmd[4] = track >> 8;			/* track number 0-0xff ?  */
			cmd[5] = track & 0xff;
			cmd[7] = res_len >> 8;
			cmd[8] = res_len & 0xff;
			cmd[9] = 0;
			errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, res, res_len);
			if (errno) {
				perror("While reading track info");
				break;
			}
			data_length		= res[1] | (res[0]  << 8);
			// track_number		= res[2] | (res[32] << 8);	  /* why? */
			session			= res[3] | (res[33] << 8);
			// is_track_mode	= res[5] & 15;
			// is_copy		= res[5] & 16;
			// is_damage		= res[5] & 32;
			// is_fixed_packet	= res[6] & 16;
			// is_packet_or_inc	= res[6] & 32;
			// is_blank		= res[6] & 64;
			// is_reserved		= res[6] & 128;
			// is_data_mode		= res[6] & 15;
			nwa_valid		= res[7] & 1;
			// lra_valid		= res[7] & 2;
			track_start		= res[11] | (res[10]<<8) | (res[ 9]<<16) | (res[ 8]<<24);
			next_writable_addr	= res[15] | (res[14]<<8) | (res[13]<<16) | (res[12]<<24);
			// free_blocks		= res[19] | (res[18]<<8) | (res[17]<<16) | (res[16]<<24);
			// packet_size		= res[23] | (res[22]<<8) | (res[21]<<16) | (res[20]<<24);
			track_size		= res[27] | (res[26]<<8) | (res[25]<<16) | (res[24]<<24);
			// last_recorded_addr	= res[32] | (res[30]<<8) | (res[29]<<16) | (res[28]<<24);
			track_end = track_start + track_size;

			if (data_length <= 30) session &= 0xff;

			disc->session_start[session-1] = MIN(disc->session_start[session-1], track_start);
			disc->session_end  [session-1] = MAX(disc->session_end  [session-1], track_end);
			if (nwa_valid) disc->next_writable[session-1] = next_writable_addr;
		};
		if (disc->session_start[disc->num_sessions-1] == INT_MAX) {
			if (disc->disc_state == DISC_STATE_FULL) {
				if (disc->last_session_state == SESSION_STATE_COMPLETE) disc->num_sessions--;
			};
		};

		return 0;
	}

	/* Start with legacy CD-ROM .... trying to get as much info from it as possible */
	disc->recordable         = 0;
	disc->rewritable         = 0;
	disc->sector_size	 = CD_SECTOR_LENGTH;
	disc->disc_state         = DISC_STATE_FULL;
	disc->last_session_state = SESSION_STATE_COMPLETE;

	/* Read session range */
	cmd[0] = 0x43;			/* READ TOC/PMA/ATAP INFO	*/
	cmd[1] = 0;			/* no times please but LBA's	*/
	cmd[2] = 1;			/* format 1 this time		*/
	cmd[6] = 0;			/* obligatory zero		*/
	cmd[7] = res_len >> 8;
	cmd[8] = res_len & 0xff;
	cmd[9] = 0;
	errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, res, res_len);
	if (errno) {
		perror("TOC reading of sessions failed");
		return errno;
	}
	first_session = read_cd_hex2(res[2]);
	last_session  = read_cd_hex2(res[3]);

	disc->num_sessions = last_session - first_session + 1;

	/* Initialise the sessions to be taken as overspanning tracks */
	for(session = 0; session <= disc->num_sessions; session++) {
		disc->session_start[session] = INT_MAX;
		disc->session_end  [session] = 0;
	}

	/* Read the complete TOC and extract information */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x43;			/* READ TOC/PMA/ATAP INFO	*/
	cmd[1] = 0;			/* no times please but LBA's	*/
	cmd[2] = 2;			/* format 2; full TOC		*/
	cmd[6] = first_session;		/* start at first session	*/
	cmd[7] = res_len >> 8;
	cmd[8] = res_len & 0xff;
	cmd[9] = 0;
	errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, res, res_len);
	if (errno) {
		perror("TOC reading of full TOC failed");
		return errno;
	}

	data_length =res[1] | (res[0] << 8);
	pos = 4;
	track_start = INT_MAX; track_end = 0;

	while (pos < data_length + 2) {
		session = read_cd_bcd(res[pos+ 0]);
		cntrl   = res[pos+1] & 15;
		addr    = res[pos+1] >> 4;
		tno     = read_cd_bcd(res[pos+ 2]);
		point   = read_cd_bcd(res[pos+ 3]);
		min     = read_cd_bcd(res[pos+ 4]);
		sec     = read_cd_bcd(res[pos+ 5]);
		frame   = read_cd_bcd(res[pos+ 6]);
		pmin    = read_cd_bcd(res[pos+ 8]);
		psec    = read_cd_bcd(res[pos+ 9]);
		pframe  = read_cd_bcd(res[pos+10]);

		/* extract information; explicit writeout. See SCSI docs */
		if (tno == 0 && session && addr == 1) {
			switch (point) {
				case 0xa0 : first_track = pmin; break;
				case 0xa1 : last_track  = pmin; break;
				case 0xa2 :
					track_end   = cd_msf2lba(pmin, psec, pframe);
					disc->session_end  [session-1] = MAX(disc->session_end  [session-1], track_end);
					break;
				default   :
					track_start = cd_msf2lba(pmin, psec, pframe);
					disc->session_start[session-1] = MIN(disc->session_start[session-1], track_start);
					break;
			};
		};
		if (tno == 0 && session && addr == 5) {
			if (point == 0xb0) {
				next_writable_addr = cd_msf2lba(min, sec, frame);
				disc_size = cd_msf2lba(pmin, psec, pframe);
				printf("UDF: ignoring B0 Q channel : next writeable address; pre MMC3 device; fix me\n");
			}

		};
		pos += 11;
	};

	/* Last session information is notoriously flawed in TOC format so update it */
	bzero(cmd, SCSI_CMD_LEN);
	cmd[0] = 0x25;		/* CD READ RECORDED CAPACITY */
	errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, res, 8);
	if (errno) {
		perror("Can't read CD recorded capacity; last session's length can be too big");
		return 0;
	};

	session = disc->num_sessions-1;
	disc->session_end[session] = res[3] | (res[2]<<8) | (res[1]<<16) | (res[0]<<24);
	return 0;
}


int udf_open_disc(char *dev_name, struct udf_discinfo **discptr) {
	struct udf_discinfo *disc;
	struct stat stat;

	if (!discptr) return EINVAL;
	*discptr = NULL;

	/* init hashtable when its not there */
	if (!udf_mountables) {
		udf_mountables = calloc(1, sizeof(struct hash_table));
		if (!udf_mountables) return ENOMEM;

		errno = hash_init_name_table(udf_mountables, 32);		/* XXX silly amount XXX */
		if (errno) return errno;
	};

	/* determine what kind of file/device we are dealing with */
	disc = calloc(1, sizeof(struct udf_discinfo));
	if (!disc) return ENOMEM;

	/* fill in the name */
	disc->dev_name = strdup(dev_name);
	if (!disc->dev_name) {
		free(disc);
		return ENOMEM;
	};

	disc->fhandle = open(dev_name, O_RDWR | O_NONBLOCK, 0);	/* no create */
	if (disc->fhandle<0) {
		perror("Failure to open device or file");
		free(disc);
		return ENODEV;
	};

	if (fstat(disc->fhandle, &stat) < 0) {
		perror("Can't stat device or file");
		close(disc->fhandle);
		free(disc);
		return ENODEV;
	};

	/* return the pointer */
	*discptr = disc;

	/* get disc info */
	return udf_get_disc_info(disc);
}


int udf_close_disc(struct udf_discinfo *disc) {
	if (!disc) return 0;
	close(disc->fhandle);
	return 0;
}


/******************************************************************************************
 *
 * Filename space conversion
 *
 ******************************************************************************************/

void udf_to_unix_name(char *result, char *id, int len, struct charspec *chsp) {
	uint16_t  raw_name[1024], unix_name[1024];
	uint16_t *pos, ch;
	uint8_t	 *opos;
	int       ucode_chars, nice_uchars;

	opos = result;
	if ((chsp->type == 0) && (strcmp(chsp->inf, "OSTA Compressed Unicode") == 0)) {
		*raw_name = *unix_name = 0;
		ucode_chars = udf_UncompressUnicode(len, id, raw_name);
		ucode_chars = MIN(ucode_chars, UnicodeLength((unicode_t *) raw_name));
		nice_uchars = UDFTransName(unix_name, raw_name, ucode_chars);
		for (pos = unix_name; nice_uchars>0; pos++, nice_uchars--) {
			ch = *pos;
			if (ch == ' ') ch = '_';	/* XXX temp hack XXX */
			*opos++ = ch & 255;
			if (!ch) break;
		};
		*opos++ = 0;
	} else {
		/* assume 8bit with length byte ? latin-1 */
		strncpy(result, id+1, strlen(id+1));
	};
}


static char *udf_get_compound_name(struct udf_mountpoint *mountpoint) {
	static char         compound[128+128+32+32+1];
	struct charspec    *charspec;
	struct udf_log_vol *udf_log_vol;
	struct udf_pri_vol *udf_pri_vol;
	char               *unix_name;

	udf_log_vol = mountpoint->udf_log_vol;
	udf_pri_vol = udf_log_vol->primary;

	charspec = &udf_pri_vol->pri_vol->desc_charset;
	assert(charspec->type == 0);
	assert(strcmp(charspec->inf, "OSTA Compressed Unicode")==0);

	unix_name = compound;

	udf_to_unix_name(unix_name, udf_pri_vol->pri_vol->volset_id, 128, charspec);
	strcat(unix_name, ":");
	unix_name += strlen(unix_name);

	udf_to_unix_name(unix_name, udf_pri_vol->pri_vol->vol_id, 32, charspec);
	strcat(unix_name, ":");
	unix_name += strlen(unix_name);
	
	udf_to_unix_name(unix_name, udf_log_vol->log_vol->logvol_id, 128, charspec);
	strcat(unix_name, ":");
	unix_name += strlen(unix_name);

	udf_to_unix_name(unix_name, mountpoint->fileset_desc->fileset_id, 32, charspec);

	return compound;
}


/******************************************************************************************
 *
 * Dump helpers for printing out information during parse
 *
 ******************************************************************************************/

void udf_dump_long_ad(char *prefix, struct long_ad *adr) {
	printf("%s at sector %d within part space %d for %d bytes\n", prefix,
			udf_rw32(adr->loc.lb_num), udf_rw16(adr->loc.part_num),
			udf_rw32(adr->len)
			);
}


void udf_dump_id(char *prefix, int len, char *id, struct charspec *chsp) {
	uint16_t  raw_name[1024];
	uint16_t *pos, ch;
	int       ucode_chars;

	if (prefix) printf("%s ", prefix);
	if ((chsp->type == 0) && (strcmp(chsp->inf, "OSTA Compressed Unicode") == 0)) {
		/* print the identifier using the OSTA compressed unicode */
		printf("`");
		ucode_chars = udf_UncompressUnicode(len, id, raw_name);
		for (pos = raw_name; ucode_chars > 0; pos++, ucode_chars--) {
			ch = *pos;
			if (!ch) break;
			if ((ch < 32) || (ch > 255)) {
				printf("[%d]", ch);
			} else {
				printf("%c", ch & 255);
			};
		};
		printf("`");
	} else {
		printf("(roughly) `%s`", id+1);
	};
	if (prefix) printf("\n");
}


void udf_dump_volume_name(char *prefix, struct udf_log_vol *udf_log_vol) {
	if (prefix) printf("%s%s", prefix, udf_log_vol->primary->udf_session->session_offset?" (local) ":" ");
	udf_dump_id(NULL, 128, udf_log_vol->primary->pri_vol->volset_id, &udf_log_vol->primary->pri_vol->desc_charset);
	printf(":");
	udf_dump_id(NULL,  32, udf_log_vol->primary->pri_vol->vol_id, &udf_log_vol->primary->pri_vol->desc_charset);
	printf(":");
	udf_dump_id(NULL, 128, udf_log_vol->log_vol->logvol_id, &udf_log_vol->log_vol->desc_charset);
	if (prefix) printf("\n");
}


/******************************************************************************************
 *
 * Sector readers
 *
 ******************************************************************************************/

int udf_read_phys(udf_sector *result, struct udf_discinfo *disc, off_t sector, char *what) {
	scsicmd		cmd;
	int		size, size_read, sect_mult;

	/* read one UDF sector at a physical address specified in UDF_SECTOR units */
	size = sizeof(udf_sector);
	size_read = 0;
	bzero(result, size);	/* just in case */

	assert(disc->sector_size);
	sect_mult = UDF_SECTOR_SIZE / disc->sector_size;
	sector *= sect_mult;
	assert(sect_mult * disc->sector_size == UDF_SECTOR_SIZE);

	switch (disc->devdrv_class & UDF_DEVDRV_CLASS) {
		case UDF_DEVDRV_CLASS_CD  :
		case UDF_DEVDRV_CLASS_DVD :
			/*
			 * Use a SCSI command to read it so we can go past the last session;
			 * allthough READ (12) is available, some older CD-ROM
			 * devices only want to do READ (10).
			 */
			bzero(cmd, SCSI_CMD_LEN);
			cmd[0] = 0x28;		/* READ (10) command		*/
			cmd[1] = 0;		/* normal access		*/
			cmd[2] = (sector >> 24) & 0xff;
			cmd[3] = (sector >> 16) & 0xff;
			cmd[4] = (sector >>  8) & 0xff;
			cmd[5] = (sector      ) & 0xff;
			cmd[6] = 0;		/* reserved */
			cmd[7] = 0;		/* MSB transfer */
			cmd[8] = sect_mult;	/* number of logical block(s); one UDF_SECTOR at a time for now */
			cmd[9] = 0;		/* control */
			errno = scsi_call(SCSI_READCMD, disc->fhandle, cmd, 10, result, size);

			if (errno) return ENOENT;
			size_read = sect_mult * disc->sector_size;	/* definition */
			break;
		default :
			/* XXX first and only reference to {p}read() XXX */
			size_read = pread(disc->fhandle, result, size, sector*UDF_SECTOR_SIZE);
			break;
	};

	if (size_read <= 0) {
UDF_VERBOSE_MAX(
		if (what) {
			printf("Can't read sector %d for %s\n", (int) sector, what);
		}
);
		if (size_read == 0) return ENOENT;
		return errno;
	};
	DEBUG(printf("%08d : %s ;read %d bytes\n", (int) sector, what, size_read));
	return 0; /* flag ok */
}


/* read one sector at a physical address using session data */
int udf_read_session_phys(udf_sector *result, struct udf_session *udf_session, off_t sector, char *what) {
	return udf_read_phys(result, udf_session->disc, sector + udf_session->session_offset, what);
}


/******************************************************************************************
 *
 * Extent readers... effectively start of UDF itself
 *
 * XXX need to be refactored for real read/write support XXX :(
 ******************************************************************************************/

/* read an physical extent with an offset; all in BYTES.... it will be written in the `result' buffer */
int udf_read_session_bytes_phys(void *result, struct udf_session *udf_session, uint64_t offset, uint64_t length, char *what) {
	udf_sector  space;
	uint8_t    *pos;
	int	    error;
	uint32_t    sector, modulo, copied;

	error = 0;
	pos = result;

	sector  = offset / UDF_SECTOR_SIZE;
	modulo  = offset % UDF_SECTOR_SIZE;

	if (modulo) printf("Reading at offset %d bytes in sector\n", modulo);

	copied = 0;
	for (; (!error) && (length > 0); sector++) {
		error = udf_read_session_phys(&space, udf_session, sector, what);
		if (error) break;

		copied = MIN(length, UDF_SECTOR_SIZE)-modulo;
		memcpy(pos, space + modulo, copied);
		modulo  = 0;
		length -= copied;
		pos    += copied;
	};
	return error;
}


/* read an extent in logical volume space at an offset in bytes for a length in bytes */
int udf_read_logvol_space(struct udf_log_vol *udf_log_vol, uint32_t vpart_num, uint32_t lb_num, uint64_t offset, uint64_t length, char *what, void **blob) {
	struct udf_volumeset	 *udf_volumeset;
	struct udf_partition	 *udf_partition;
	struct udf_session	 *udf_session;
	struct udf_part_mapping  *udf_part_mapping;
	struct spare_map_entry	 *sp_entry;

	uint32_t		  lb_size;
	void			**part_blob;	/* only used in META */
	uint32_t		  part_num, part_start, part_length;
	uint64_t		  eff_addr, eff_sector, eff_offset;
	uint64_t		  trans_addr, trans_sector, trans_length;
	uint64_t		  res_length, real_length;
	uint8_t			 *blob_pos;
	uint32_t		  entry, entries;

	uint64_t		  packet_sector, packet_offset;
	uint32_t		  packet_len;

	uint32_t		  vat_entries, *vat_pos;

	/* bounds checking for mapped partition number */
	if (vpart_num > udf_log_vol->num_part_mappings) {
		printf("\t\t\tPartition number %d outside bounds!\n", vpart_num);
		return EFAULT;
	};

	DEBUG(
		printf("Read logvol space for %s, from vpart %d, lb_num %d at offset %d, for %d bytes\n",
			what, vpart_num, lb_num, (int) offset, (int) length)
	);
	/* map the requested partition map to the udf partition */
	assert(udf_log_vol->part_mappings);
	udf_part_mapping = udf_log_vol->part_mappings;
	while (udf_part_mapping && udf_part_mapping->udf_virt_part_num != vpart_num)
		udf_part_mapping = udf_part_mapping->next;

	assert(udf_part_mapping);	/* call me paranoid */
	part_num = udf_part_mapping->udf_phys_part_num;

	/* get the partition */
	udf_volumeset = udf_log_vol->primary->volumeset;
	udf_partition = udf_volumeset->parts;
	while (udf_partition && (udf_rw16(udf_partition->partition->part_num) != part_num))
		udf_partition = udf_partition->next;

	if (!udf_partition) {
		printf("\t\t\tNo information known about partition %d yet!\n", part_num);
		printf("\t\t\t\tPlease insert volume %d of this volumeset and try again\n", udf_part_mapping->vol_seq_num);
		return ENOENT;
	};

	lb_size     = udf_rw32(udf_log_vol->log_vol->lb_size);
	part_start  = udf_rw32(udf_partition->partition->start_loc);
	part_length = udf_rw32(udf_partition->partition->part_len);

	/* get the offset (in bytes) in the partition and check its validity */
	offset  += lb_num * lb_size;
	if (offset + length > ((uint64_t) part_start + part_length) * UDF_SECTOR_SIZE) {
		printf("\t\t\t%s is outside partition %d!\n", what, part_num);
		return EFAULT;
	};

	/* XXX Is this like a kernel module? XXX */
	*blob = malloc(length);
	if (!*blob) {
		printf("\t\t\tOut of memory claiming %s memory\n", what);
		return ENOMEM;
	};

	udf_session = udf_partition->udf_session;
	/* do the address translations based on the partition mapping type */
	switch (udf_part_mapping->udf_part_mapping_type) {
		case UDF_PART_MAPPING_PHYSICAL :
			/* nothing to be done; physical is logical */
			eff_addr = (uint64_t) part_start*UDF_SECTOR_SIZE + offset;
			// printf("DEBUG: read phys eff_addr = %ld, eff sector %ld, partnr = %d, part_start = %d for %ld bytes\n",
			//	eff_addr, (eff_addr)/2048, part_num, part_start, length);
			return udf_read_session_bytes_phys(*blob, udf_session, eff_addr, length, what);

		case UDF_PART_MAPPING_VIRTUAL :
			vat_entries = udf_part_mapping->vat_entries;
			vat_pos = (uint32_t *) udf_part_mapping->vat_translation;

			/* this translation is dependent on partition relative sector numbers */
			eff_sector = offset / UDF_SECTOR_SIZE;
			eff_offset = offset - eff_sector * UDF_SECTOR_SIZE;

			blob_pos = *blob;
			do {
				trans_sector = vat_pos[eff_sector];
				DEBUG(
					printf("VAT: translating eff_sector %d (eff_offset=%d) to sector %d\n",
						(uint32_t) eff_sector, (uint32_t) eff_offset, (uint32_t) trans_sector)
				);
				trans_addr   = (part_start + trans_sector)*UDF_SECTOR_SIZE + eff_offset;
				trans_length = MIN(length, UDF_SECTOR_SIZE-eff_offset);

				/* get this translated piece */
				errno = udf_read_session_bytes_phys(blob_pos, udf_session, trans_addr, trans_length, what);
				blob_pos      += trans_length;
				length        -= trans_length;
				eff_offset     = 0;
				eff_sector    += 1;
			} while (length && !errno);
			return errno;

		case UDF_PART_MAPPING_SPARABLE :
			/* transform on packet-length base */
			packet_len    = udf_rw16(udf_part_mapping->udf_pmap->pms.packet_len);
			entries = udf_rw32(udf_part_mapping->sparing_table->rt_l);

			eff_addr = (uint64_t) part_start*UDF_SECTOR_SIZE + offset;
			packet_sector = packet_len * (eff_addr / (packet_len * UDF_SECTOR_SIZE));
			packet_offset = eff_addr - (packet_sector * UDF_SECTOR_SIZE);	/* within packet */

			blob_pos = *blob;
			do {
				/* translate this packet */
				trans_sector = packet_sector;
				sp_entry = &udf_part_mapping->sparing_table->entries[0];
				for(entry = 0; entry < entries; entry++) {
					if (udf_rw32(sp_entry->org) - packet_sector == 0) {
						trans_sector = udf_rw32(sp_entry->map);
						break;
					};
					sp_entry++;
				};
				trans_addr   = trans_sector*UDF_SECTOR_SIZE + packet_offset;
				trans_length = MIN(length, packet_len * UDF_SECTOR_SIZE - packet_offset);
				/* get this translated piece of the packet */
				errno = udf_read_session_bytes_phys(blob_pos, udf_session, trans_addr, trans_length, what);
				blob_pos      += trans_length;
				length        -= trans_length;
				packet_offset  = 0;
				packet_sector += packet_len;
			} while (length && !errno);
			return errno;

			break;
		case UDF_PART_MAPPING_META :
			errno = udf_read_file_part(udf_part_mapping->meta_file, offset, length, blob, &res_length, &real_length);
			if (errno) return errno;

			blob_pos  = ((uint8_t *) *blob) + res_length;
			offset   += res_length;
			length	 -= res_length;
			part_blob = NULL;
			while (length) {
				errno = udf_read_file_part(udf_part_mapping->meta_file, offset, length, part_blob, &res_length, &real_length);
				if (errno) {
					free(*blob);
					free(*part_blob);
					return errno;
				};
				memcpy(blob_pos, *part_blob, res_length);
				free(*part_blob);

				offset   += res_length;
				length   += res_length;
				blob_pos += res_length;
			};
			break;
		case UDF_PART_MAPPING_ERROR :
		default :
			break;
	};
	printf("Unsupported or bad mapping %d; can't translate\n", udf_part_mapping->udf_part_mapping_type);
	free(*blob);
	*blob = NULL;

	return EFAULT;
}



int udf_get_icb_struct_length(int addr_type) {
	if (addr_type == UDF_ICB_SHORT_ALLOC) return sizeof(struct short_ad);
	if (addr_type == UDF_ICB_LONG_ALLOC)  return sizeof(struct long_ad);
	return 0;
}


int udf_lookup_node(struct udf_log_vol *udf_log_vol, struct long_ad *addr, struct udf_icb_node **nodeptr) {
	struct udf_icb_node *node;
	uint32_t vpart_num, lb_num, length;
	int32_t  error;

	assert(nodeptr);
	*nodeptr = NULL;

	hash_lookup_entry(&udf_log_vol->udf_icb_nodes, (void *) addr, (void **) nodeptr);
	if (*nodeptr) return 0;

	/* ok ... load node and set up a new record */
	node = calloc(1, sizeof(struct udf_icb_node));
	if (!node) {
		printf("UDF: out of memory claiming ICB node\n");
		return ENOMEM;
	};

	node->udf_log_vol = udf_log_vol;
	node->location    = *addr;
	node->dscrptr     = NULL;

	vpart_num = udf_rw16(addr->loc.part_num);
	lb_num    = udf_rw32(addr->loc.lb_num);
	length    = udf_rw32(addr->len);
	error = udf_read_logvol_space(udf_log_vol, vpart_num, lb_num, 0, length, "ICB node", (void **) &node->dscrptr);
	if (error) {
		free(node);
		return error;
	};
	error = hash_add_entry(&udf_log_vol->udf_icb_nodes, (void *) &node->location, (void *) node);
	if (error) {
		free(node->dscrptr);
		free(node);
		return ENOMEM;
	};
	*nodeptr = node;
	return 0;
}


/*
 * Major function that reads a part from an offset in a file described by a
 * udf_icb_node structure. It gives an extent back that is either the requested
 * length or the space remained in the extent descriptor that covers the start
 * of the blob at offset bytes; thus preventing having to read in multiple
 * extents. Its a recursive function for some ICB strategy's can have a depth
 * that is need to be known.
 * Its main use is by udf_read_file_part() that should be used instead of
 * calling this function directly.
 */
static int udf_read_file_part_recursion(struct udf_icb_node *udf_icb_node,
		uint64_t offset, uint64_t req_length, int depth,
		void **blob, uint64_t *result_length, uint64_t *return_offset, uint64_t *real_length) {
	struct udf_log_vol	*udf_log_vol;
	struct file_entry	*file_entry;
	struct udf_icb_node	*sub_node;
	union dscrptr     *dscrptr;
	union icb	  *icb;
	uint8_t		  *pos;
	uint32_t	   data_length, retrieve_length;
	uint32_t	   len, lb_num, vpart_num;
	uint32_t	   icb_len;
	uint8_t		   addr_type;
	uint32_t	   entries, prev_direct_entries;
	uint32_t	   direct_entries, indirect_entries;
	uint16_t	   strategy, strat_param16;
	uint64_t           curr_offset;
	int32_t		   error;

	assert(udf_icb_node);

	if (!blob || !result_length) return EINVAL;

	*blob          = NULL;
	*result_length = 0;

	/* see what we got here */
	dscrptr = udf_icb_node->dscrptr;
	udf_log_vol = udf_icb_node->udf_log_vol;

	errno = udf_check_tag(dscrptr);
	if (errno) return errno;
	errno = udf_check_tag_payload(dscrptr);
	if (errno) return errno;

	if (udf_rw16(dscrptr->tag.id) != TAGID_FENTRY) {
		printf("read_file_part : help! i got a %d tag\n", udf_rw16(udf_icb_node->dscrptr->tag.id));
		udf_dump_descriptor(udf_icb_node->dscrptr);
		return EFAULT;
	};

	file_entry = &dscrptr->fe;
	prev_direct_entries = udf_rw32(file_entry->icbtag.prev_num_dirs);
	strat_param16  = udf_rw16(* (uint16_t *) (file_entry->icbtag.strat_param));
	entries        = udf_rw16(file_entry->icbtag.max_num_entries);
	strategy       = udf_rw16(file_entry->icbtag.strat_type);
	data_length    = udf_rw32(file_entry->l_ad);
	addr_type      = udf_rw16(file_entry->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;

	/*
	 * get the extent in wich the file is recorded... due to strategy
	 * types this might result in recursive ICB tree parsing. :( The
	 * `advantage' is that whatever the form of the tree the reading of
	 * the tree can be done strategy type independent. This algorithm is
	 * however not specified in the standard. The sequence of the blocks
	 * in the file is left recursive descent as far as I could determine.
	 */
	/* XXX CLEAN ME UP! XXX */
	switch (strategy) {
		case	1:
		case	2:
		case	3:
			return ENODEV;	/* yet! */
		case 	4:
			/* just this direct entries ICB; no indirect entries */
			if (entries > 1)
				printf("UDF warning : strategy 4 with more than 1 entry\n");
			/* XXX confusion : only direct entries are allowed but why is `entries' == 1 ? independent of splitting? XXX */
			direct_entries   = data_length; // entries;
			indirect_entries = 0;
			break;
		case 4096:
			/* one linked list of entries : one direct the other indirect */
			if (strat_param16 != 1)
				printf("UDF warning : strategy 4096 with stat param != 1 : assuming number of direct entries\n");
			direct_entries   = strat_param16;
			indirect_entries = 1;
			break;
		default :
			return ENODEV; /* not even defined yet */
	};

	curr_offset = 0;
	pos = &file_entry->data[0] + udf_rw32(file_entry->l_ea);
	icb_len = 0;

	if (real_length)     *real_length = udf_rw64(file_entry->inf_len);
	if (req_length == 0)  req_length  = udf_rw64(file_entry->inf_len);

#if 0
	if (offset > udf_rw64(file_entry->inf_len)) {
		return EINVAL;
	};
#endif
	if ((addr_type == UDF_ICB_INTERN_ALLOC) || (data_length == 0)) {
		/* all data is internal to this node : copy it out */
		if (offset > data_length) {
			printf("UDF %s file offset out of bounds\n", data_length?"internal":"empty");
			return EINVAL;
		};
		data_length = MIN(req_length, data_length);
		*blob = malloc(data_length-offset);
		if (!*blob) return ENOMEM;

		memcpy(*blob, pos, data_length-offset);
		*result_length = data_length;
		return 0;
	};

	/* first process direct entries */
	while (direct_entries) {
		icb  = (union icb *) pos;
		vpart_num = udf_rw16(udf_icb_node->location.loc.part_num);
		lb_num = 0;
		len    = 0;
		switch (addr_type) {
			/* XXX VIOLATES ecma167 48.14.1.1 XXX */
			case UDF_ICB_SHORT_ALLOC  :
				len       = udf_rw32(icb->s_ad.len);
				lb_num    = udf_rw32(icb->s_ad.lb_num);
				icb_len   = sizeof(struct short_ad);
				break;
			case UDF_ICB_LONG_ALLOC   :
				len       = udf_rw32(icb->l_ad.len);
				vpart_num = udf_rw16(icb->l_ad.loc.part_num);
				lb_num    = udf_rw32(icb->l_ad.loc.lb_num);
				icb_len   = sizeof(struct long_ad);
				break;
			case UDF_ICB_EXT_ALLOC    :
				printf("UDF extended allocators not supported nor defined in ICB\n");
				return ENODEV;
			case UDF_ICB_INTERN_ALLOC :
				printf("Sanity check : I can't be here\n");
				exit(1);
		};
		if (offset < len) {
			/* found an extent that fits */
			data_length = MIN(req_length, len-offset);
			retrieve_length = data_length;

			/* we get a blob back from udf_read_logvol_space */
			error = udf_read_logvol_space(udf_log_vol, vpart_num, lb_num, offset, retrieve_length, "File", blob);
			if (error) {
				free(*blob);
				*blob = NULL;
				*result_length = 0;
				return error;
			};
			*result_length = retrieve_length;
			return 0;
		};

		offset        -= len;
		pos           += icb_len;
		data_length   -= icb_len;
		direct_entries-= icb_len;
	};
	if (offset >= 0) {
		printf("UDF: WARNING file is truncated; missing %d bytes\n", (int) req_length);
		*blob = NULL;
		*result_length = 0;
		return 0;
	}

	/* process following indirect entries -> pointing to a new ICB */
	printf("UDF: WARNING Indirect entries not implemented yet; please contact author!\n");
	sub_node = NULL;
#if 0
	while (data_length) {
		icb  = (union icb *) pos;
		error = udf_lookup_node(udf_log_vol, &icb->l_ad, &sub_node);
		if (!error) {
			error  = udf_read_file_part_recursion(sub_node, offset, req_length, depth+1, blob, result_length, return_offset, real_length);
		};
		if (error) perror("UDF: Error reading indirect ICB entry");
		if (error) return EIO;

		len = *return_offset - offset;
		offset      -= len;
		pos         += icb_len;
		data_length -= icb_len;
	};
#endif

	if (depth == 0) return EINVAL;

	*return_offset = offset;
	return EFAULT;
}


int udf_read_file_part(struct udf_icb_node *udf_icb_node, uint64_t offset, uint64_t length, void **blob, uint64_t *result_len, uint64_t *real_len) {
	uint64_t result_offset;

	errno = udf_read_file_part_recursion(udf_icb_node, offset, length, 0, blob, result_len, &result_offset, real_len);
	if (errno == EINVAL) {
		printf("UDF File offset out of bounds\n");
	};
	return errno;
}


int udf_read_complete_file(struct udf_icb_node *udf_icb_node, void **blobptr, uint64_t *lengthptr) {
	uint8_t		*blob, *blob_piece;
	uint64_t	 length, got_length, real_length;

	blob = NULL;
	got_length = length = real_length = 0;
	do {
		/* XXX we can't do realloc() in kernel ... move to uio structs one day! XXX */
		errno = udf_read_file_part(udf_icb_node, length, real_length-length, (void **) &blob_piece, &got_length, &real_length);
		if (errno) break;
		blob = realloc(blob, length + got_length);
		if (blob) {
			memcpy(blob + length, blob_piece, got_length);
			length += got_length;
		};
		free(blob_piece);
	} while (length < real_length && got_length != 0);

	if (errno) {
		free(blob);
		return errno;
	};
	*blobptr   = blob;
	*lengthptr = length;
// XXX for extreme debugging only XXX 
#if 0
{
	int fh;
	fh = open("grab_blob", O_CREAT | O_WRONLY, 0666);
	write(fh, blob, length);
	close(fh);
}
#endif

	return 0;
}


/******************************************************************************************
 *
 * UDF logic
 *
 ******************************************************************************************/

/* not used extensively enough yet */
int udf_check_tag(union dscrptr *dscr) {
	struct desc_tag *tag = &dscr->tag;
	uint8_t *pos, sum, cnt;

	/* check TAG header checksum */
	pos = (uint8_t *) tag;
	sum = 0;

	for(cnt = 0; cnt < 16; cnt++) {
		if (cnt != 4) sum += *pos;
		pos++;
	};
	if (sum != tag->cksum) {
		/* bad tag header checksum; this is not a valid tag */
		return EINVAL;
	};
	return 0;
}


int udf_check_tag_payload(union dscrptr *dscr) {
	struct desc_tag *tag = &dscr->tag;
	uint16_t crc;

	/* check payload CRC if applicable */
	if (udf_rw16(tag->desc_crc_len) == 0) return 0;

	crc = udf_cksum(((uint8_t *) tag) + UDF_DESC_TAG_LENGTH, udf_rw16(tag->desc_crc_len));
	if (crc != udf_rw16(tag->desc_crc)) {
		/* bad payload CRC; this is a broken tag */
		return EINVAL;
	};

	return 0;
}


int udf_check_tag_presence(union dscrptr *dscr, int TAG) {
	struct desc_tag *tag = &dscr->tag;

	errno = udf_check_tag(dscr);
	if (errno) return errno;

	if (udf_rw16(tag->id) != TAG) {
		DEBUG(fprintf(stderr, "looking for tag %d but found %d\n", TAG, udf_rw16(tag->id)));
		return ENOENT;
	};

	return 0;
}


/* for malloc() purposes ... rather have an upperlimit than an exact size */
int udf_calc_tag_malloc_size(union dscrptr *dscr) {
	int32_t size, tag_id;

	tag_id = udf_rw16(dscr->tag.id);

	switch (tag_id) {
		case TAGID_LOGVOL :
			size = sizeof(struct logvol_desc) + udf_rw32(dscr->lvd.mt_l);
			break;
		case TAGID_UNALLOC_SPACE :
			size  = sizeof(struct unalloc_sp_desc);
			size += udf_rw32(dscr->usd.alloc_desc_num) * sizeof(struct extent_ad);
			break;
		case TAGID_FID :
			size = UDF_FID_SIZE + dscr->fid.l_fi + udf_rw16(dscr->fid.l_iu);
			size = (size + 3) & ~3;
			break;
		case TAGID_LOGVOL_INTEGRITY :
			size  = sizeof(struct logvol_int_desc) - sizeof(uint32_t) + udf_rw32(dscr->lvid.l_iu);
			size += (2 * udf_rw32(dscr->lvid.num_part) * sizeof(uint32_t));
			break;
		default :
			size = sizeof(union dscrptr);
			break;
	};
	return size;
}


struct udf_volumeset *udf_search_volumeset(char *volset_id) {
	struct udf_volumeset *here;

	here = udf_volumeset_list;
	while (here) {
		/* if its empty, we recycle it */
		if (!here->primaries) return here;

		assert(here->primaries->pri_vol);
		if (memcmp(here->primaries->pri_vol->volset_id, volset_id, 128) == 0) return here;
		here = here->next;
	};
	return NULL;
}


struct udf_pri_vol *udf_search_primary(struct udf_volumeset *set, char *id) {
	struct udf_pri_vol *here;

	here = set->primaries;
	while (here) {
		assert(here->pri_vol);
		assert(*here->pri_vol->vol_id ==  *id);
		if (memcmp(here->pri_vol->vol_id, id, 32) == 0) return here;
		here = here->next;
	};
	return NULL;
}


struct udf_log_vol *udf_search_logical_volume_in_primary(struct udf_pri_vol *primary, char *logvol_id) {
	struct udf_log_vol *here;

	here = primary->log_vols;
	while (here) {
		assert(here->log_vol);
		assert(*here->log_vol->logvol_id == *logvol_id);
		if (memcmp(here->log_vol->logvol_id, logvol_id, 128) == 0) return here;
		here = here->next;
	};
	return NULL;
}


/* hash table functions */
int udf_icb_node_compare(void *v1, void *v2) {
	struct long_ad *n1, *n2;

	n1 = (struct long_ad *) v1;
	n2 = (struct long_ad *) v2;

	/* XXX sanity checking ? i.e. is it pointing to the same but different? XXX */
	return memcmp(&n1->loc, &n2->loc, sizeof(struct lb_addr));
}


uint32_t udf_icb_node_calc_hash(void *v1) {
	struct long_ad *n1;

	n1 = (struct long_ad *) v1;

	/* for `normal' filingsystems this will even be unique */
	return udf_rw32(n1->loc.lb_num) + (udf_rw16(n1->loc.part_num) << 24);
}


/* called not very often ... ; free incomming when not needed */
int udf_proc_pri_vol(struct udf_session *udf_session, struct udf_pri_vol **current, struct pri_vol_desc *incomming) {
	struct udf_volumeset *volset;
	struct udf_pri_vol   *primary;

	volset = udf_search_volumeset(incomming->volset_id);
	if (!volset) {
		/* create a volume set */
		volset = calloc(1, sizeof(struct udf_volumeset));
		if (!volset) {
			free(incomming);
			return ENOMEM;
		};

		/* populate and link in */
		volset->primaries   = NULL;		/* just in case */
		volset->parts       = NULL;		/* just in case */
		volset->max_partnum = 0;
		volset->next        = udf_volumeset_list;
		udf_volumeset_list  = volset;
	};
	assert(volset);

	primary = udf_search_primary(volset, incomming->vol_id);
	if (!primary) {
		/* create a primary volume */
		primary = calloc(1, sizeof(struct udf_pri_vol));
		if (!primary) {
			free(incomming);
			return ENOMEM;
		};

		/* populate and link in */
		primary->pri_vol     = incomming;
		primary->udf_session = udf_session;
		primary->volumeset   = volset;

		primary->next = volset->primaries;
		volset->primaries = primary;

		*current = primary;
		return 0;
	};

	/* mark as current */
	*current = primary;

	/* ok ... we now need to check if this new descriptor is a newer version */
	if (udf_rw32(incomming->seq_num) <= udf_rw32(primary->pri_vol->seq_num)) {
		if (udf_session->session_num <= (*current)->udf_session->session_num) {
			DEBUG(printf("UDF: DISCARDING primary descriptor for its the same but higher session number\n"));
			/* its an older one; ignore */
			free(incomming);
			return 0;
		}
	};
	DEBUG(printf("UPDATING primary descriptor for it has a higher session number\n"));

	/* update the primary volume descriptor */
	free(primary->pri_vol);
	primary->pri_vol = incomming;
	primary->udf_session = udf_session;

	return 0;
}


/* not called often ; free lvid when not used */
int udf_proc_logvol_integrity(struct udf_log_vol *udf_log_vol, struct logvol_int_desc *lvid) {
	errno = udf_check_tag((union dscrptr *) lvid);
	if (errno) {
		free(lvid);
		return errno;	/* return error on faulty tag */
	};

	/* getting a terminator tag is an OK condition */
	if (udf_rw16(lvid->tag.id) == TAGID_TERM) {
		free(lvid);
		return 0;
	};

	/* not getting an logical volume itegrity volume descriptor is an error now */
	if (udf_rw16(lvid->tag.id) != TAGID_LOGVOL_INTEGRITY) {
		free(lvid);
		return EBADF;	/* XXX error code? XXX */
	};

	/* check CRC on the contents of the logvol integrity */
	errno = udf_check_tag_payload((union dscrptr *) lvid);
	if (errno) {
		free(lvid);
		return errno;
	};

	/* replace existing logical integrity volume if applicable */
	if (udf_log_vol->logvol_integrity) {
		UDF_VERBOSE_TABLES(printf("Replacing logical volume integrity with a newer version\n\n"));
		free(udf_log_vol->logvol_integrity);
	};
	udf_log_vol->logvol_integrity = lvid;

	return 0;
}


int udf_proc_logvol_integrity_sequence(struct udf_log_vol *udf_log_vol) {
	union dscrptr	*dscr;
	struct udf_logvol_info *impl;
	uint32_t	 sector, length, lvid_len, num_sectors;

	sector = udf_rw32(udf_log_vol->log_vol->integrity_seq_id.loc);
	length = udf_rw32(udf_log_vol->log_vol->integrity_seq_id.len);

	while (length) {
		dscr = calloc(1, sizeof(udf_sector));
		if (!dscr) break;

		errno = udf_read_session_phys((udf_sector *) dscr, udf_log_vol->primary->udf_session, sector, "logical volume integrity descriptor (LVID)");
		if (!errno) errno = udf_check_tag(dscr);
		if (!errno) errno = udf_check_tag_payload(dscr);
		if (errno) {
			free(dscr);
			break;
		};

		lvid_len = udf_calc_tag_malloc_size(dscr);
		if (lvid_len > UDF_SECTOR_SIZE) {
			free(dscr);
			dscr = calloc(1, lvid_len);
			if (!dscr) break;

			errno = udf_read_session_bytes_phys(dscr, udf_log_vol->primary->udf_session, sector * UDF_SECTOR_SIZE, lvid_len, "long LIVD");
			if (!errno) errno = udf_check_tag(dscr);
			if (!errno) errno = udf_check_tag_payload(dscr);
			if (errno) {
				free(dscr);
				break;
			};
		};

		UDF_VERBOSE_MAX(udf_dump_descriptor(dscr));

		errno = udf_proc_logvol_integrity(udf_log_vol, &dscr->lvid);
		if (errno) break;
		if (udf_rw16(dscr->tag.id) == TAGID_TERM) break;

		num_sectors = (lvid_len + UDF_SECTOR_SIZE-1) / UDF_SECTOR_SIZE;
		length -= num_sectors * UDF_SECTOR_SIZE;
		sector += num_sectors;

		if (udf_rw32(dscr->lvid.next_extent.len)) {
			sector = udf_rw32(dscr->lvid.next_extent.loc);
			length = udf_rw32(dscr->lvid.next_extent.len);
		};
	};

	if (!udf_log_vol->logvol_integrity) {
		fprintf(stderr, "UDF: Fatal error; no logical volume integrity descriptor found; create a new one\n");
		printf("XXX fix me XXX\n");
		exit(1);
	};
	if (errno && (errno != EIO)) {
		printf("UDF: Warning, error while reading logical volume integrity tables; assuming dirty volume\n");
		udf_log_vol->logvol_integrity->integrity_type = udf_rw32(UDF_INTEGRITY_OPEN);
	};

	/* process information contained in logical volume integrity descriptor */
	impl = (struct udf_logvol_info *) (udf_log_vol->logvol_integrity->tables + 2*udf_rw16(udf_log_vol->logvol_integrity->num_part));
	udf_log_vol->min_udf_readver  = udf_rw16(impl->min_udf_readver);
	udf_log_vol->min_udf_writever = udf_rw16(impl->min_udf_writever);
	udf_log_vol->max_udf_writever = udf_rw16(impl->max_udf_writever);
	/* XXX fileId ? free space counters ? num files? num dirs? */

	return 0;
}


/* not called often; free incomming when not needed */
int udf_proc_log_vol(struct udf_pri_vol *primary, struct logvol_desc *incomming) {
	struct udf_log_vol	 *logical;
	struct udf_part_mapping	 *part_mapping, *next_part_mapping;
	union  udf_pmap		 *pmap;
	uint32_t		  part_cnt, pmap_type, pmap_size;
	char			 *pmap_pos;

	logical = udf_search_logical_volume_in_primary(primary, incomming->logvol_id);
	if (!logical) {
		/* create a logical volume */
		logical = calloc(1, sizeof(struct udf_log_vol));
		if (!logical) {
			free(incomming);
			return ENOMEM;
		};

		/* populate and link in */
		hash_init_generic_table(&logical->udf_icb_nodes, 100, udf_icb_node_compare, udf_icb_node_calc_hash);

		logical->primary  = primary;
		logical->next     = primary->log_vols;
		primary->log_vols = logical;
	} else {
		/* ok ... we now need to check if this new descriptor is a newer version */
		if (udf_rw32(incomming->seq_num) < udf_rw32(logical->log_vol->seq_num)) {
			/* its an older one; ignore */
			free(incomming);
			return 0;
		};
	};

	/* update the logical volume descriptor and its mappings; first delete old partition mappings allocated before */
	part_mapping = logical->part_mappings;
	while (part_mapping) {
		next_part_mapping = part_mapping->next;
		/*
			free(part_mapping->sparing_table);
			free(part_mapping->vat_file_entry);
			free(part_mapping->vat);
			free(part_mapping->meta_file);
			free(part_mapping->meta_mirror_file);
			free(part_mapping->meta_bitmap_file);
		*/
		free(part_mapping);
		part_mapping = next_part_mapping;
	};
	logical->part_mappings = NULL;

	/* use the new logical volume and preprocess it */
	free(logical->log_vol);
	logical->log_vol = incomming;

	/* process the logical volume integrity sequence */
	udf_proc_logvol_integrity_sequence(logical);

	/* build up the partion mappings */
	logical->num_part_mappings = udf_rw32(incomming->n_pm);

	pmap_pos = &logical->log_vol->maps[0];
	for(part_cnt = 0; part_cnt < logical->num_part_mappings; part_cnt++) {
		/* get a new part_mapping structure */
		part_mapping = calloc(1, sizeof(struct udf_part_mapping));
		assert(part_mapping);		/* XXX check with partition mapping destructor etc XXX */

		/* add to list */
		part_mapping->next = logical->part_mappings;
		logical->part_mappings = part_mapping;

		/* process */
		pmap = (union udf_pmap *) pmap_pos;
		pmap_type = pmap->data[0];
		pmap_size = pmap->data[1];

		part_mapping->udf_virt_part_num = part_cnt;
		part_mapping->udf_pmap = pmap;
		switch (pmap_type) {
			case 1:
				part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_PHYSICAL;
				part_mapping->vol_seq_num = udf_rw16(pmap->pm1.vol_seq_num);
				part_mapping->udf_phys_part_num = udf_rw16(pmap->pm1.part_num);
				break;
			case 2:
				part_mapping->vol_seq_num = udf_rw16(pmap->pm2.vol_seq_num);
				part_mapping->udf_phys_part_num = udf_rw16(pmap->pm2.part_num);
				if (strncmp(pmap->pm2.part_id.id, "*UDF Virtual Partition", UDF_REGID_ID_SIZE) == 0) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_VIRTUAL;
					break;
				};
				if (strncmp(pmap->pm2.part_id.id, "*UDF Sparable Partition", UDF_REGID_ID_SIZE) == 0) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_SPARABLE;
					break;
				};
				if (strncmp(pmap->pm2.part_id.id, "*UDF Metadata Partition", UDF_REGID_ID_SIZE) == 0) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_META;
					fprintf(stderr, "WARNING: untested `metadata' partition type encountered\n");
					break;
				};
				printf("HELP ... found unsupported type 2 partition mapping id `%s`; marking broken\n", pmap->pm2.part_id.id);
			default:
				part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
		};
		switch (part_mapping->udf_part_mapping_type) {
			case UDF_PART_MAPPING_PHYSICAL :
			case UDF_PART_MAPPING_ERROR :
			case UDF_PART_MAPPING_VIRTUAL :
				/* logical volume integrity warning is not usefull for these types */
				break;
			default :
				if (udf_rw32(logical->logvol_integrity->integrity_type) != UDF_INTEGRITY_CLOSED) {
					printf("UDF: Warning, logical volume is not closed!\n");
				};
				break;
		};

		pmap_pos += pmap_size;		/* variable length array :( */
	};
	return 0;
}


/* not called often; free incomming when not needed */
int udf_proc_part(struct udf_pri_vol *primary, struct part_desc *incomming) {
	struct udf_partition *udf_partition;
	struct udf_volumeset *udf_volset;
	uint32_t new_part_num;

	assert(primary);
	assert(primary->pri_vol);

	udf_volset = udf_search_volumeset(primary->pri_vol->volset_id);
	assert(udf_volset);

	new_part_num = udf_rw16(incomming->part_num);

	/* look if we allready got it */

	udf_partition = udf_volset->parts;
	while (udf_partition && (udf_rw16(udf_partition->partition->part_num) != new_part_num))
		udf_partition = udf_partition->next;

	/* we have space... now check if this is a newer one than the one known */
	if (udf_partition) {
		if (udf_rw32(incomming->seq_num) < udf_rw32(udf_partition->partition->seq_num)) {
			/* its an older version */
			free(incomming);
			return 0;
		};
	} else {
		/* get us a new udf_partition */
		udf_partition = calloc(1, sizeof(struct udf_partition));
		if (!udf_partition) {
			free(incomming);
			return ENOMEM;
		};
		/* link it in */
		udf_partition->next = udf_volset->parts;
		udf_volset->parts = udf_partition;
		udf_volset->max_partnum = MAX(udf_volset->max_partnum, new_part_num);
	};
	assert(udf_partition);

	/* copy this new partition descriptor in the list */
	if (udf_partition->partition) free(udf_partition->partition);
	udf_partition->partition   = incomming;
	udf_partition->udf_session = primary->udf_session;

	return 0;
}


/* not called often; free incomming when not needed */
int udf_proc_filesetdesc(struct udf_log_vol *udf_log_vol, struct fileset_desc *incomming) {
	struct udf_mountpoint *mp;

	if (udf_rw16(incomming->tag.id) != TAGID_FSD) {
		printf("IEEE! Encountered a non TAGID_FSD in this fileset descriptor sequence!!!\n");
		free(incomming);
		return EFAULT;
	};

	/* lookup fileset descriptor in this logical volume; interestingly fileset_num is KEY! */
	mp = udf_log_vol->mountpoints;
	while (mp && udf_rw32(mp->fileset_desc->fileset_num) != udf_rw32(incomming->fileset_num)) {
		mp = mp->next;
	};

	if (!mp) {
		/* add a new mountpoint! */
		mp = calloc(1, sizeof(struct udf_mountpoint));
		if (!mp) {
			free(incomming);
			return ENOMEM;
		};
		mp->udf_log_vol  = udf_log_vol;

		mp->next = udf_log_vol->mountpoints;
		udf_log_vol->mountpoints = mp;
	} else {
		/* update mountpoint? */
		if (udf_rw32(incomming->fileset_desc_num) <= udf_rw32(mp->fileset_desc->fileset_desc_num)) {
			/* we got a newer one */
			free(incomming);
			return 0;
		};
		/* delete from mountables list */
		hash_delete_entry(udf_mountables, mp->mount_name);
		free(mp->mount_name);
		free(mp->fileset_desc);
	};

	/* create new `mount_name' and add it to the mountables list */
	mp->fileset_desc = incomming;
	mp->mount_name = strdup(udf_get_compound_name(mp));
	hash_add_entry(udf_mountables, mp->mount_name, mp);

	return 0;
}


int udf_retrieve_volume_space(struct udf_discinfo *disc, struct udf_session *udf_session, struct extent_ad *extent) {
	struct udf_pri_vol *udf_pri_vol;
	union dscrptr	*dscr;
	uint32_t	 sector, length, dscr_len, num_sectors;
	int		 tag_id;

	udf_pri_vol = NULL;

	sector = udf_rw32(extent->loc);
	length = udf_rw32(extent->len);
	while (length) {
		dscr = calloc(1, sizeof(udf_sector));
		if (!dscr) break;

		errno = udf_read_session_phys((udf_sector *) dscr, udf_session, sector, "volume descriptor");
		if (!errno) errno = udf_check_tag(dscr);
		if (!errno) errno = udf_check_tag_payload(dscr);
		if (errno) {
			free(dscr);
			break;
		}

		dscr_len = udf_calc_tag_malloc_size(dscr);
		if (dscr_len > UDF_SECTOR_SIZE) {
			free(dscr);
			dscr = calloc(1, dscr_len);
			if (!dscr) break;

			errno = udf_read_session_bytes_phys(dscr, udf_session, sector * UDF_SECTOR_SIZE, dscr_len, "long volume descriptor");
			if (!errno) errno = udf_check_tag(dscr);
			if (!errno) errno = udf_check_tag_payload(dscr);
		};

		tag_id = udf_rw16(dscr->tag.id);
		num_sectors = (dscr_len + UDF_SECTOR_SIZE-1) / UDF_SECTOR_SIZE;

		/* proc volume descriptor starting at sector `volume_sector' */
		UDF_VERBOSE_MAX(udf_dump_descriptor(dscr));
		switch (tag_id) {
			case TAGID_PRI_VOL       :
				errno = udf_proc_pri_vol(udf_session, &udf_pri_vol, &dscr->pvd);
				break;
			case TAGID_PARTITION     :
				errno = udf_proc_part(udf_pri_vol, &dscr->pd);
				break;
			case TAGID_LOGVOL        :
				errno = udf_proc_log_vol(udf_pri_vol, &dscr->lvd);
				break;
			case TAGID_TERM          :
				free(dscr);
				return 0;	/* terminator */
			case TAGID_UNALLOC_SPACE :
				/* unallocated space descriptor */
				/* Specifies space that is not claimed yet in partitions (!) */
				UDF_VERBOSE(fprintf(stderr, "UDF: Unhandled `unallocated space descriptor'; ignored\n"));
				break;
			case TAGID_IMP_VOL       :
				/* implemenation use volume descriptor */
				/* Specifies information about the implementator */
				UDF_VERBOSE_MAX(fprintf(stderr, "UDF: `implementation use volume descriptor' ignored\n"));
				break;
			case TAGID_VOL           :
				fprintf(stderr, "UDF : untested volume space extender encountered\n");
				break;
			default :
				printf("XXX Unhandled volume sequence %d; freeing\n", tag_id);
				free(dscr);
				break;
		};

		length -= num_sectors * UDF_SECTOR_SIZE;
		sector += num_sectors;

		if (tag_id == TAGID_VOL) {
			sector = udf_rw32(dscr->vdp.next_vds_ex.loc);
			length = udf_rw32(dscr->vdp.next_vds_ex.len);
			free(dscr);
		};
	};

	return errno;
}


int udf_check_for_vat(struct udf_log_vol *udf_log_vol, struct udf_part_mapping *part_mapping, struct file_entry *file_entry) {
	struct udf_icb_node	 vat_node;
	struct regid 		*regid;
	uint64_t		 vat_length;
	uint32_t		*vat_pos;
	uint8_t			*vat, *pos;

	/*
	 * OK ... now get the VAT itself; better set up a dummy node for the file entry to simplify
	 * matters since the VAT can be fragmented too (!)
	 */

	vat_node.udf_log_vol = udf_log_vol;
	vat_node.location.loc.part_num = udf_rw16(part_mapping->udf_pmap->pmv.part_num);
	vat_node.location.loc.lb_num   = 0;
	vat_node.location.len          = 0;
	vat_node.dscrptr = (union dscrptr *) file_entry;

	errno = udf_read_complete_file(&vat_node, (void **) &vat, &vat_length);
	if (errno) return errno;

	part_mapping->vat_file_entry = file_entry;
	part_mapping->vat            = (struct udf_vat *) vat;
	part_mapping->vat_length     = vat_length;

	/*XXX move to be a part_mapping_quirk ? XXX */
	if (file_entry->icbtag.file_type == UDF_ICB_FILETYPE_VAT) {
		/* we are in UDF 2.00+ userland */
		part_mapping->vat_translation = &(part_mapping->vat->data[0]) + udf_rw16(part_mapping->vat->impl_use_len);
		part_mapping->vat_entries = (udf_rw16(part_mapping->vat->header_len) - udf_rw16(part_mapping->vat->impl_use_len))/4;	/* definiton */
	} else {
		/* still in the old UDF 1.50 and older userland; update? its notoriously broken */
		vat_length = udf_rw64(file_entry->inf_len);	/* hack hack  */
		part_mapping->vat_entries = (vat_length-36)/4;	/* definition */
		part_mapping->vat_length = vat_length;
		part_mapping->vat_translation = vat;
		part_mapping->vat = NULL;
	};

	/* check this VAT for validity */
	if (file_entry->icbtag.file_type != UDF_ICB_FILETYPE_VAT) {
		/* check the old UDF 1.50 VAT */
		DEBUG(printf("CHECK UDF 1.50 VAT\n"));
		pos = part_mapping->vat_translation;
		vat_pos = (uint32_t *) pos;
		regid = (struct regid *) (vat_pos+part_mapping->vat_entries);
		if (strncmp(regid->id, "*UDF Virtual Alloc Tbl", 22) == 0) return 0;
		DEBUG(printf("\tFAULT\n"));
		return ENOENT;
	};
	return 0;
}


int udf_retrieve_supporting_tables(struct udf_log_vol *udf_log_vol) {
	struct udf_part_mapping  *part_mapping;
	struct udf_session	 *udf_session;
	struct udf_sparing_table *sparing_table;
	struct file_entry	 *file_entry;
	struct long_ad		  location;
	uint8_t			 *vat;
	uint32_t		  spar_loc, spar_size;
	uint64_t		  vat_loc, last_vat_loc;
	uint32_t		  vat_part;
	int			  part_num, spar_num;
	int			  session_num;

	/*
	 * if there are any virtual or sparable partition in this logical
	 * volume, try to find their supporting tables so we can find the rest
	 */
	part_mapping = udf_log_vol->part_mappings;
	while (part_mapping) {
		part_num = part_mapping->udf_virt_part_num;
		UDF_VERBOSE_TABLES(printf("\tFor partition mapping %d->%d\n", part_num, part_mapping->udf_phys_part_num));
		switch (part_mapping->udf_part_mapping_type) {
			case UDF_PART_MAPPING_ERROR    :
				/* nothing to be done for these */
				break;
			case UDF_PART_MAPPING_PHYSICAL :
				/* nothing to be done for these; no supporting tables */
				break;
			case UDF_PART_MAPPING_VIRTUAL  :
				/*
				 * we have to find a good VAT at the END of the session. Since VAT's are
				 * only to be used on WORM's and need to written as last, the strategy is
				 * to go for the predicted end of this session and walk UP
				 */
				udf_session = udf_log_vol->primary->udf_session;
				session_num = udf_session->session_num;

				UDF_VERBOSE_TABLES(printf("\t\tSearching for the VAT :"));
				if (udf_session->session_length == 0) {
					UDF_VERBOSE(
						printf("\t\tThis virtual partition is inaccessible since its its size is not known;\n");
						printf("\t\tTry to insert the disc in a CD or DVD recordable device to access it.\n");
					);
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
					continue;
				};

				vat = NULL; vat_part = 0;
				if (udf_session->disc->next_writable[session_num]) {
					vat_loc = udf_session->disc->next_writable[session_num];
				} else {
					vat_loc = udf_session->disc->session_end[session_num];
				}
				last_vat_loc = vat_loc + 7;
				vat_loc -= 4; /* give some extra slack since sizes are not allways given up correctly */

				file_entry = malloc(UDF_SECTOR_SIZE);
				if (!file_entry) {
					printf("IEE!! out of memory for VAT file entry\n");
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
					udf_log_vol->broken = 1;
					continue;
				};
				assert(file_entry);

				/* try to find the fileid for the VAT */
				do {
					DEBUG(
						printf("Trying VAT at sector %d in session\n", (int) vat_loc)
					);
					errno = udf_read_session_bytes_phys(file_entry, udf_log_vol->primary->udf_session,
							vat_loc * UDF_SECTOR_SIZE, UDF_SECTOR_SIZE, "VAT file entry");
					if (!errno) errno = udf_check_tag_presence((union dscrptr *) file_entry, TAGID_FENTRY);
					if (!errno) errno = udf_check_tag_payload( (union dscrptr *) file_entry);
					if (!errno) errno = udf_check_for_vat(udf_log_vol, part_mapping, file_entry);
					if (!errno) {
						break;
					} else {
						vat_loc++;
						if (vat_loc > last_vat_loc) errno = EIO;
					};
				} while (errno != EIO);
				if (errno) {
					printf("IEE!! I was looking for a VAT but didnt found it\n");
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
					udf_log_vol->broken = 1;
					continue;
				};
				UDF_VERBOSE_TABLES(printf("found VAT file-entry at device logical sector %d\n", (uint32_t) vat_loc));
				UDF_VERBOSE_TABLES(printf("\t\tFound %d byte VAT descriptor+table\n", (uint32_t) part_mapping->vat_length));

				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) file_entry));
				UDF_VERBOSE_MAX(udf_dump_vat_table(part_mapping));
				/* XXX process updated information about num files/dirs, UDF read/write versions etc. XXX */
				if (part_mapping->vat) {
					/* update UDF versions */
					udf_log_vol->min_udf_readver  = udf_rw16(part_mapping->vat->min_udf_readver);
					udf_log_vol->min_udf_writever = udf_rw16(part_mapping->vat->min_udf_writever);
					udf_log_vol->max_udf_writever = udf_rw16(part_mapping->vat->max_udf_writever);
				};
				break;
			case UDF_PART_MAPPING_SPARABLE :
				/* we have to find a good sparing table; address are in device logical blocks */
				sparing_table = NULL;
				spar_size = udf_rw32(part_mapping->udf_pmap->pms.st_size);
				sparing_table = malloc(spar_size);
				if (!sparing_table) {
					printf("UDF: Out of memory claiming sparing table\n");
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
					continue;
				};

				for(spar_num = 0; spar_num < part_mapping->udf_pmap->pms.n_st; spar_num++) {
					spar_loc = udf_rw32(part_mapping->udf_pmap->pms.st_loc[spar_num]);
					if (!sparing_table) break;

					/* fetch spar_loc's table ; on THIS session. */
					errno = udf_read_session_bytes_phys(sparing_table, udf_log_vol->primary->udf_session,
							spar_loc * UDF_SECTOR_SIZE, spar_size, "Sparing table");
					if (!errno) errno = udf_check_tag_presence((union dscrptr *) sparing_table, TAGID_SPARING_TABLE);
					if (!errno) {
						UDF_VERBOSE_TABLES(printf("\t\tFound the sparing table\n"));
						part_mapping->sparing_table = sparing_table;
						break;
					};
				};
				if (!part_mapping->sparing_table) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
				};
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->sparing_table));
				break;
			case UDF_PART_MAPPING_META :
				/*
				 * set up common locator parts; the files are located inside the `part_num' partion where
				 * this partition is a added layer on.
				 */
				printf("XXX WARNING : reading `supporting tables' for a metadata partition; code not tested yet\n");
				location.loc.part_num = udf_rw16(part_mapping->udf_pmap->pmm.part_num);
				location.len          = UDF_SECTOR_SIZE;		/* defined as maximum size */

				UDF_VERBOSE_TABLES(printf("Reading metadata partition filedescriptor\n"));
				location.loc.lb_num   = udf_rw32(part_mapping->udf_pmap->pmm.meta_file_lbn);
				udf_lookup_node(udf_log_vol, &location, &part_mapping->meta_file);
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->meta_file));

				UDF_VERBOSE_TABLES(printf("Reading metadata partition mirror filedescriptor\n"));
				location.loc.lb_num   = udf_rw32(part_mapping->udf_pmap->pmm.meta_mirror_file_lbn);
				udf_lookup_node(udf_log_vol, &location, &part_mapping->meta_mirror_file);
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->meta_mirror_file));

				UDF_VERBOSE_TABLES(printf("Reading metadata partition allocation bitmap filedescriptor\n"));
				location.loc.lb_num   = udf_rw32(part_mapping->udf_pmap->pmm.meta_bitmap_file_lbn);
				udf_lookup_node(udf_log_vol, &location, &part_mapping->meta_bitmap_file);
				UDF_VERBOSE_TABLES(udf_dump_descriptor((union dscrptr *) part_mapping->meta_bitmap_file));

				/* if something is wrong, then mark it as a broken partition */
				if (!(part_mapping->meta_file || part_mapping->meta_mirror_file) || !part_mapping->meta_bitmap_file) {
					part_mapping->udf_part_mapping_type = UDF_PART_MAPPING_ERROR;
				};
				break;
		};
		part_mapping = part_mapping->next;
	};
	UDF_VERBOSE_TABLES(printf("\n"));
	return 0;
}


int udf_retrieve_fileset_descriptor(struct udf_log_vol *udf_log_vol) {
	struct long_ad		*fsd_loc;
	struct fileset_desc	*new_fsd;
	uint32_t		 part_num, lb_num, length;
	int32_t			 error;

	error = 0; /* flag OK */

	fsd_loc   = &udf_log_vol->log_vol->_lvd_use.fsd_loc;
	part_num  = udf_rw16(fsd_loc->loc.part_num);
	lb_num    = udf_rw32(fsd_loc->loc.lb_num);
	length    = udf_rw32(fsd_loc->len);

	while (length && !error) {
		UDF_VERBOSE_TABLES(
			printf("\tFileset descriptor extent remaining at sector %d within partion %d for %d bytes\n", lb_num, part_num, length)
		);

		/* only go for ONE fsb at a time */
		error = udf_read_logvol_space(udf_log_vol, part_num, lb_num, 0, UDF_SECTOR_SIZE, "Fileset descriptor", (void **) &new_fsd);
		if (!error) error = udf_check_tag((union dscrptr *) new_fsd);
		if (error || !new_fsd || (new_fsd && (udf_rw16(new_fsd->tag.id) == TAGID_TERM))) {
			/* end of sequence */
			UDF_VERBOSE_TABLES(
				printf("\t\t(Terminator) ");
				if (!new_fsd || error) printf("; unrecorded"); else printf("; explicit");
				printf("\n");
			);
			if (new_fsd) free(new_fsd);
			break;
		};

		UDF_VERBOSE_MAX(udf_dump_descriptor((union dscrptr *) new_fsd));
		udf_proc_filesetdesc(udf_log_vol, new_fsd);

		if (udf_rw32(new_fsd->next_ex.len) == 0) {
			/* next entry */
			lb_num += 1;
			length -= UDF_SECTOR_SIZE;
		} else {
			/* follow the next extent */
			fsd_loc  = &new_fsd->next_ex;
			part_num = udf_rw16(fsd_loc->loc.part_num);
			lb_num   = udf_rw32(fsd_loc->loc.lb_num);
			length   = udf_rw32(fsd_loc->len);
		};
	};
	UDF_VERBOSE_TABLES(printf("\n"));

	return error;
}


/*
 * udf_eliminate_predescessor_volumesets()
 *
 * We are faced with a courious problem : we are to examine the partitions and
 * determine which are successors of eachother.  This is propably most
 * relevant only on WORM media though. We could consider the following rules :
 * 	1) `glue' according to strict UDF level 1 rules ?
 * 	2) use heuristics info i.e. NERO, DirectCD and mkisofs quirks ?
 * 	3) use overlapping partitions to detect relationships ?
 *
 * Using 1 would imply no multi-volume discs and thus glue everything but that
 * could easily be wrong. Selecting by volumeset names is not possible for
 * Nero f.e. just creates random volumeset names every session and uses no
 * volume version information and thus also violates the UDF rules.
 *
 * Using 2 would be tricky; we know a few programs but what if more are
 * developped? We then would be at loss.
 *
 * Using 3 would imply some calculation but is fail-safe in both supporting
 * multiple volumes on one disc (they seperate) and in supporting
 * multi-session WORM media for these will refer to eachother. Offcource NERO
 * could faul this by just extending the zero partion to the whole disc in its
 * ignorance and thus create false overlapping over other independent
 * sessions. This is to be investigated. I dont know how NERO will react on
 * this situation.
 *
 * Propably method 3 would be good to try :
 *
 * Follow the disc and check for all sessions in order to mark the ones with
 * overlapping partitions as `inactive' and keep the latest one active.
 * Sessions with the `local' quirk are seperate allmost by default; should
 * be change the offsets? would not be too difficult but possible.
 */

void udf_eliminate_predescessor_volumesets(struct udf_discinfo *disc) {
	struct udf_volumeset	*anc_vol_set;
	struct udf_volumeset	*sib_vol_set;
	struct udf_partition	*anc_part;
	struct udf_partition	*sib_part;
	int			 anc_partnum;
	int			 sib_partnum;
	uint32_t		 anc_start, anc_end;
	uint32_t		 sib_start, sib_end;
	uint32_t		 overlap_start, overlap_end;
	uint32_t		 anc_session;
	uint32_t		 sib_session;

	anc_vol_set = udf_volumeset_list;
	while (anc_vol_set) {
		sib_vol_set = anc_vol_set->next;
		while (sib_vol_set) {
			/* compare these two volume sets but only process partitions on _this_ disc */
DEBUG(
printf("checking volset %s with volset %s\n", anc_vol_set->primaries->pri_vol->volset_id+1, sib_vol_set->primaries->pri_vol->volset_id+1
));
			anc_part = anc_vol_set->parts;
			while (anc_part) {
				if (anc_part->udf_session->disc != disc) {
					anc_part = anc_part->next;
					continue;
				};

				anc_session = anc_part->udf_session->session_num;
				anc_start = 0;
				if (disc->session_quirks[anc_session] & CD_SESS_QUIRK_SESSION_LOCAL)
					anc_start += disc->session_start[anc_session];
				anc_start += udf_rw32(anc_part->partition->start_loc);
				anc_end    = anc_start + udf_rw32(anc_part->partition->part_len);

				sib_part = sib_vol_set->parts;
				while (sib_part) {
					if (sib_part->udf_session->disc != disc) {
						sib_part = sib_part->next;
						continue;
					};
					sib_session = sib_part->udf_session->session_num;
#if 1
					/* can `session local' volumes even be considered part/successor ? */
					if (disc->session_quirks[sib_session] & CD_SESS_QUIRK_SESSION_LOCAL) {
						sib_part = sib_part->next;
						continue;
					};
#endif
					sib_start = 0;
					if (disc->session_quirks[sib_session] & CD_SESS_QUIRK_SESSION_LOCAL)
						sib_start += disc->session_start[sib_session];
					sib_start += udf_rw32(sib_part->partition->start_loc);
					sib_end    = sib_start + udf_rw32(sib_part->partition->part_len);
DEBUG(
anc_partnum = udf_rw16(anc_part->partition->part_num);
sib_partnum = udf_rw16(sib_part->partition->part_num);
printf("\t\tchecking partition %d with partition %d ([%d-%d] x [%d-%d])\n", anc_partnum, sib_partnum, anc_start, anc_end, sib_start, sib_end)
);
					overlap_start = MAX(sib_start, anc_start);
					overlap_end   = MIN(sib_end,   sib_end);
					if (overlap_start < overlap_end)  {
DEBUG(
printf("\t\t\tOVERLAP!\n")
);
						if (sib_session < anc_session) {
							/* the sibbling is older */
			UDF_VERBOSE_TABLES(
				printf("\tVolume set ");
				udf_dump_id(NULL, 128, anc_vol_set->primaries->pri_vol->volset_id, &anc_vol_set->primaries->pri_vol->desc_charset);
				printf(" is a newer version of volume set ");
				udf_dump_id(NULL, 128, sib_vol_set->primaries->pri_vol->volset_id, &sib_vol_set->primaries->pri_vol->desc_charset);
				printf("\n");
			);
							sib_vol_set->obsolete = 1;
							break;
						};
					}; /* overlap */
					if (sib_vol_set->obsolete) break;
					sib_part = sib_part->next;
				}; /* sibling partition */
				if (sib_vol_set->obsolete) break;
				anc_part = anc_part->next;
			}; /* ancestor partition */
			sib_vol_set = sib_vol_set->next;
		}; /* sibling volume set */
		anc_vol_set = anc_vol_set->next;
	}; /* ancestor volume set */
}



int udf_get_anchors(struct udf_discinfo *disc) {
	struct udf_session	*udf_session, *last_session;
	udf_sector		 sector;
	union dscrptr		*dscr = (union dscrptr *) sector;
	uint32_t		 session_start, session_end;
	int			 session, error;

	/* Get all anchors */
	disc->sessions = NULL;
	udf_session = last_session = NULL;
	for(session = 0; session < disc->num_sessions; session++) {
		/* check for anchors ; no volume recognition data ? */
		session_start = disc->session_start[session];
		session_end   = disc->session_end  [session];
		error = udf_read_phys(&sector, disc, session_start + 256, "Anchor");
		if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
		if (error) {
			error = udf_read_phys(&sector, disc, session_end - 256, "Anchor");
			if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
			if (error) {
				error = udf_read_phys(&sector, disc, session_end, "Anchor");
				if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
				if (error) {
					/* special recordable case; not likely to happen */
					error = udf_read_phys(&sector, disc, session_start + 512, "Anchor");
					if (!error) error = udf_check_tag_presence(dscr, TAGID_ANCHOR);
				};
			}
		}
		if (!error) {
			udf_session = calloc(1, sizeof(struct udf_session));
			assert(udf_session);

			udf_session->anchor = dscr->avdp;
			udf_session->disc   = disc;
			udf_session->session_num = session;
			udf_session->session_offset   = 0;
			udf_session->session_length   = disc->session_end[session] - disc->session_start[session];
			disc->session_quirks[session] = 0;

			/* detect quirks */
			if ((disc->session_start[session] > 0) && (udf_session->anchor.main_vds_ex.loc == 32)) {
				disc->session_quirks[session] |= CD_SESS_QUIRK_SESSION_LOCAL;
				udf_session->session_offset = disc->session_start[session];
			};

			/* add to tail of session list */
			if (last_session)  last_session->next = udf_session;
			if (!last_session) disc->sessions = udf_session;
	
			last_session = udf_session;

			disc->num_udf_sessions++;
		};
		/* record status of this volume */
		disc->session_is_UDF[session] = error ? 0 : 1;
	};
	return 0;
}


int udf_get_volumeset_space(struct udf_discinfo *disc) {
	struct udf_session *udf_session;
	int error;

	/* Rip all volume spaces */
	UDF_VERBOSE(printf("Retrieving Volume information\n"));
	udf_session = disc->sessions;
	while (udf_session) {
		UDF_VERBOSE_MAX(printf("Session %d volumes : \n", udf_session->session_num));

		error = udf_retrieve_volume_space(disc, udf_session, &udf_session->anchor.main_vds_ex);
		if (error) {
			printf("\nRetrieving session %d's reserve volume information due to error\n", udf_session->session_num);
			error = udf_retrieve_volume_space(disc, udf_session, &udf_session->anchor.reserve_vds_ex);
		};

		udf_session = udf_session->next;
	};
	UDF_VERBOSE_MAX(printf("\n\n"));
	return 0;
}


int udf_get_logical_volumes_info(struct udf_discinfo *disc) {
	struct udf_volumeset	*udf_volumeset;
	struct udf_pri_vol	*udf_pri_vol;
	struct udf_log_vol	*udf_log_vol;

	udf_volumeset = udf_volumeset_list;
	while (udf_volumeset) {
		if (!udf_volumeset->obsolete) {
			udf_pri_vol = udf_volumeset->primaries;
			while (udf_pri_vol) {
				if (udf_pri_vol->udf_session->disc == disc) {
					udf_log_vol = udf_pri_vol->log_vols;
					while (udf_log_vol) {
						/* supporting tables */
						UDF_VERBOSE(udf_dump_volume_name("Retrieving support tables for logical volume", udf_log_vol));
						errno = udf_retrieve_supporting_tables(udf_log_vol);
	
						/* fileset descriptor */
						UDF_VERBOSE(udf_dump_volume_name("Retrieving fileset for logical volume", udf_log_vol));
						if (!errno) errno = udf_retrieve_fileset_descriptor(udf_log_vol);
	
						if (errno) udf_log_vol->broken = 1;
						udf_log_vol = udf_log_vol->next;
					}; /* logical */
				}; /* disc */
				udf_pri_vol = udf_pri_vol->next;
			}; /* primary */
		}; /* if */
		udf_volumeset = udf_volumeset->next;
	}; /* volumeset */

	return 0;
};


static void udf_process_session_range(struct udf_discinfo *disc, int *enabled, int low, int high) {
	int session;

	if (!disc) return;

	high = MIN(high, disc->num_sessions-1);
	session = low;

	for (session = low; session <= high; session++) {
		enabled[session] = 1;
	};
}


/* range is specified in -3,5,7 or 5-6,8- etc */
static int udf_process_session_range_string(struct udf_discinfo *disc, char *range) {
	struct udf_session *udf_session, *next_udf_session, *last_ok;
	char *pos, *nop;
	int low, high, len, session;
	int enabled[MAX_SESSIONS];

	if (!range) return 0;
	DEBUG(printf("UDF range debugging string '%s'\n", range));

	if (disc) {
		/* disable all */
		for (session = 0; session < disc->num_sessions; session++) {
			enabled[session] = 0;
		};
	};

	/* parse string */
	nop = strdup(range);
	pos = range;
	if (sscanf(pos, "-%u%n%s", &high, &len, nop) >= 1) {
		DEBUG(printf("UDF range match till %d\n", high));
		udf_process_session_range(disc, enabled, 0, high);
		pos += len;
	};
	if (*pos && *pos == ',') pos++;
	while (*pos) {
		if (sscanf(pos, "%u%n%s", &low, &len, nop) >= 1) {
			pos += len;
			if (*pos == '-') {
				pos++;
				if (!*pos) {
					DEBUG(printf("UDF range match from %d\n", low));
					udf_process_session_range(disc, enabled, low, INT_MAX);
					free(nop);
					return 0;
				};
				if (sscanf(pos, "%u%n%s", &high, &len, nop) >= 1) {
					pos += len;
					DEBUG(printf("UDF range match from %d to %d\n", low, high));
					udf_process_session_range(disc, enabled, low, high);
				};
			} else {
				if (!*pos || (*pos == ',')) {
					DEBUG(printf("UDF range match %d\n", low));
					udf_process_session_range(disc, enabled, low, low);
				};
			};
			if (*pos && (*pos != ',')) {
				fprintf(stderr, "UDF range matching : ',' expected at %s\n", pos);
				free(nop);
				return ENOENT;
			};
			pos++;
		} else {
			fprintf(stderr, "UDF range matching : number expected at %s\n", pos);
			free(nop);
			return ENOENT;
		};
	};
	free(nop);

	DEBUG(printf("UDF range matching : all ok till the end\n"));
	if (!disc) return 0;

	last_ok = NULL;
	udf_session = disc->sessions;
	while (udf_session) {
		next_udf_session = udf_session->next;
		session = udf_session->session_num;
		if (!enabled[session]) {
			/* remove this session */

			fprintf(stderr, "UDF: disabling UDF session %d on request\n", session);
			if (!last_ok) disc->sessions = next_udf_session;
			if (last_ok) {
				last_ok->next = next_udf_session;
			};
			free(udf_session);
			disc->session_is_UDF[session] = 0;
		} else {
			last_ok = udf_session;
		};
		udf_session = next_udf_session;
	};

	return 0;
}


int udf_check_session_range(char *range) {
	return udf_process_session_range_string(NULL, range);
}


int udf_open_and_read_disc(char *devname, char *range, struct udf_discinfo **disc) {
	errno = udf_open_disc(devname, disc);
	if (errno) return errno;

	errno = udf_get_anchors(*disc);
	UDF_VERBOSE(udf_dump_disc_anchors(*disc));

	if (range) {
		UDF_VERBOSE(printf("Selecting UDF sessions '%s' as specified\n", range));
		udf_process_session_range_string(*disc, range);
		UDF_VERBOSE(udf_dump_disc_anchors(*disc));
	};

	/* no UDF partitions so bail out */
	if ((*disc)->num_udf_sessions == 0) return 0;

	errno = udf_get_volumeset_space(*disc);
	if (errno) return errno;

	UDF_VERBOSE(printf("Finding volumesets dependences\n"));
	udf_eliminate_predescessor_volumesets(*disc);

	UDF_VERBOSE_TABLES(udf_dump_alive_sets());

	UDF_VERBOSE(printf("Retrieving logical volumes their VAT's/Sparing tables\n"));
	errno = udf_get_logical_volumes_info(*disc);

	return errno;
}


/******************************************************************************************
 *
 * udf_state handling for writing support
 *
 ******************************************************************************************/



/******************************************************************************************
 *
 * Directory and other conversion UDF logic
 *
 ******************************************************************************************/

int udf_opendir(struct udf_icb_node *node, int offset, struct udf_readdir_stream **readdirptr) {
	struct udf_readdir_stream  *readdir;

	readdir = *readdirptr = malloc(sizeof(struct udf_readdir_stream));
	if (!readdir) return ENOMEM;

	readdir->udf_icb_node = node;
	readdir->dirpart  = NULL;
	readdir->pos      = 0;
	readdir->dir_len  = 0;
	readdir->eof      = 0;
	readdir->fileno   = 0;

	errno = udf_read_complete_file(node, (void **) (&readdir->dirpart), &readdir->dir_len);

	if (errno) {
		free(readdir);
		*readdirptr = NULL;
		return errno;
	};
	return 0;
}


int udf_readdir_entry(struct udf_readdir_stream *readdir, struct dirent *entry, struct long_ad *icb, int *type) {
	struct fileid_desc	*fid;
	struct udf_log_vol	*udf_log_vol;
	uint32_t	 entry_length;
	uint8_t		*fid_name;

	udf_log_vol = readdir->udf_icb_node->udf_log_vol;
	fid = NULL;
	fid_name = NULL;

	while (readdir->pos < readdir->dir_len) {
		fid = (struct fileid_desc *) (readdir->dirpart + readdir->pos);
		fid_name = fid->data + udf_rw16(fid->l_iu);
		/*
		 * the `dirent' entry wants a filetype! now thats difficult ... as in
		 * only dirs or so? or really go for the node and look?
		 */
	
		/* advance */
		entry_length = udf_calc_tag_malloc_size((union dscrptr *) fid);
		readdir->pos += entry_length;

		/* if its not a deleted file, return */
		if ((fid->file_char & UDF_FILE_CHAR_DEL) == 0) {
			entry->d_fileno = 1;
			entry->d_reclen = sizeof(struct dirent);
			entry->d_type   = DT_UNKNOWN;
			if (readdir->fileno) {
				udf_to_unix_name(entry->d_name, fid_name, fid->l_fi, &udf_log_vol->log_vol->desc_charset);
				if (icb)  *icb  = fid->icb;
				if (type) *type = fid->file_char;
				if (fid->file_char & UDF_FILE_CHAR_DIR) entry->d_type = DT_DIR;
				if (fid->file_char & UDF_FILE_CHAR_PAR) strcpy(entry->d_name, "..");
			} else {
				strcpy(entry->d_name, ".");
				if (icb)  *icb  = readdir->udf_icb_node->location;
			if (type) *type = UDF_FILE_CHAR_DIR;
				entry->d_type = DT_DIR;
			};
			//entry->d_namlen = strlen(entry->d_name); /* XXX */
	
			readdir->fileno++;
			return 0;
		};
	}
	readdir->eof = (readdir->pos >= readdir->dir_len);

	return 0;
}


int udf_closedir(struct udf_readdir_stream *readdir) {
	if (readdir) {
		free(readdir->dirpart);
		free(readdir);
	};
	return 0;
}

