/* $NetBSD$ */

/*
 * File "udfclient.c" is part of the UDFclient toolkit.
 * File $Id: udfclient.c,v 1.23 2003/09/14 14:19:03 imago Exp $ $Name: release-ro1 $
 *
 * Copyright (c) 2003 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udfclient/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "errno.h"
#include "udf.h"
#include "udf_bswap.h"
#include "hashtable.h"


/* switches */

//#define DEBUG(a) (a)
#define DEBUG(a) if (0) { a; };


#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))


/* include the dump parts ... in order to get a more sane splitting */
extern void udf_dump_alive_sets(void);
extern void udf_dump_file_entry_node(struct udf_icb_node *udf_icb_node);


/* globals */
extern int errno;
extern int udf_verbose;
extern int scsilib_verbose;


char		    *curdir_name;
struct udf_icb_node *curdir;


struct udfclient_dirread_stream {
	struct udf_icb_node		*curdir;
	struct udf_readdir_stream	*udf_readdir;

	struct hash_entry		*udf_mountpoint_entry;

	int eof;
};


/*
 * XXX
 * FTP client; de volumes vooraan zetten in de VFS ; evt. in meedere subdirs.
 * general file name format 
 *
 * /volset:pri:log/udfpath
 *     of
 * /volset/pri/log/udfpath/
 * 
 * XXX
 */

int udfclient_opendir(struct udf_icb_node *curdir, struct udfclient_dirread_stream **dirreadptr) {
	struct udfclient_dirread_stream *dirread;

	dirread = calloc(1, sizeof(struct udfclient_dirread_stream));
	assert(dirread);

	dirread->curdir  = curdir;
	dirread->udf_readdir = NULL;
	dirread->udf_mountpoint_entry = NULL;
	dirread->eof = 0;

	*dirreadptr = dirread;

	if (dirread->curdir) {
		return udf_opendir(dirread->curdir, 0, &dirread->udf_readdir);
	};

	/* ``rootdir'' */
	if (!udf_mountables) {
		dirread->eof = 1;
	} else {
		dirread->udf_mountpoint_entry = udf_mountables->seq_head;
		if (!dirread->udf_mountpoint_entry) dirread->eof = 1;
	};

	return 0;
}


int udfclient_dirread_entry(struct udfclient_dirread_stream *dirread, struct dirent *entry, struct udf_log_vol **udf_log_volptr, struct long_ad *icb, int *type) {
	struct udf_mountpoint *mountable;
	struct udf_log_vol    *udf_log_vol;
	int   skip;

	if (dirread->eof) return 0;
	if (dirread->curdir) {
		errno = udf_readdir_entry(dirread->udf_readdir, entry, icb, type);
		dirread->eof = dirread->udf_readdir->eof;
		return errno;
	};

	/* rootdir */
	for(;;) {
		if (!dirread->udf_mountpoint_entry) dirread->eof = 1;
		if (dirread->eof) return 0;

		mountable = (struct udf_mountpoint *) dirread->udf_mountpoint_entry->entry;
		udf_log_vol = mountable->udf_log_vol;

		skip = (udf_log_vol && udf_log_vol->broken);
		if (skip) {
			fprintf(stderr, "UDF: internal error... broken logical volume's fileset in mountables\n");
		};
		if (!skip) {
			/* build the dirent */
			strcpy(entry->d_name, mountable->mount_name);
			entry->d_type = DT_DIR;
			if (udf_log_volptr) *udf_log_volptr = udf_log_vol;
			if (icb && mountable->fileset_desc) *icb = mountable->fileset_desc->rootdir_icb;
		};
		/* advance to the next in the ``rootdir''; the state points to the first candidate */
		dirread->udf_mountpoint_entry = dirread->udf_mountpoint_entry->seq_next;

		if (!skip) return 0;
	};
}


int udfclient_closedir(struct udfclient_dirread_stream *dirread) {
	if (dirread->udf_readdir) udf_closedir(dirread->udf_readdir);
	free(dirread);
	return 0;
}


char *udfclient_realpath(char *cur_path, char *relpath) {
	char *resultpath, *here, *pos;

	resultpath = calloc(1, sizeof(cur_path) + sizeof(relpath) +1024);
	assert(resultpath);

	strcpy(resultpath, "/");
	strcat(resultpath, cur_path);
	strcat(resultpath, "/");

	/* check if we are going back to `root' */
	if (relpath && *relpath == '/') {
		strcpy(resultpath, "");
	};
	strcat(resultpath, relpath);
	/* now clean up the resulting string by removing double slashes */
	here = resultpath;
	while (*here) {
		pos = here; while (strncmp(pos, "//", 2) == 0) pos++;
		if (pos != here) strcpy(here, pos);
		here++;
	};

	/* remove '.' and '..' sequences */
	here = resultpath;
	while (*here) {
		// printf("transformed to %s\n", resultpath);
		/* check for internal /./ and trailing /. */
		if (strncmp(here, "/./", 3) == 0) {
			strcpy(here+1, here + 3);
			continue;
		};
		if (strcmp(here, "/.")==0) {
			strcpy(here+1, here + 2);
			continue;
		};
		if (strncmp(here, "/../", 4) == 0) {
			strcpy(here+1, here + 4);
			/* go for the parent */
			pos = here-1; while (*pos && *pos != '/') pos--; pos++;
			strcpy(pos, here+1);
			here = pos;
			continue;
		};
		if (strcmp(here, "/..")==0) {
			strcpy(here+1, here + 3);
			/* go for the parent */
			pos = here-1; while (*pos && *pos != '/') pos--; pos++;
			strcpy(pos, here+1);
			here = pos;
			continue;
		};
		here++;
	};

	return resultpath;
}


/* translate a name to a directory icb address */
int udfclient_lookup_name(struct udf_log_vol *udf_log_vol, struct udf_icb_node *cur_node, char *restpath,
		struct dirent *dirent, struct udf_icb_node **new_node) {
	struct udfclient_dirread_stream *dirread;
	struct long_ad location;
	char *next_element, *slashpos;

	//printf("\nLookup name entry logvol=%p, cur_node = %p, restpath = %s\n", udf_log_vol, cur_node, restpath);
	/* check if we are here */
	if (!restpath) {
		*new_node = cur_node;
		return 0;
	};

	/* extract a path element */
	//printf("Splitting `%s' in ", restpath);

	slashpos = strchr(restpath, '/');
	if (slashpos) *slashpos++ = 0;
	next_element = restpath;
	restpath = slashpos;

	if (restpath && *restpath == 0) restpath = NULL;

	//printf("`%s' slash `%s'\n", next_element, restpath);

	if (strlen(next_element) == 0 && cur_node == NULL) {
		/* we are in '/' */
		strcpy(dirent->d_name, "/");
		dirent->d_type = DT_DIR;
		// printf("In root; restpath = %s\n", restpath);
		return udfclient_lookup_name(udf_log_vol, cur_node, restpath, dirent, new_node);
	};

	//printf("\nDirectory search cur_node = %p for %s\n", cur_node, next_element);
	/* lookup the name and pass on */
	errno = udfclient_opendir(cur_node, &dirread);
	if (errno) {
		perror("udfclient_opendir failed :");
	} else {
		// int udfclient_dirread_entry(void *cookie, struct dirent *entry, struct long_ad *icb, int *type) {
		udfclient_dirread_entry(dirread, dirent, &udf_log_vol, &location, NULL);
		while (!dirread->eof) {
			//printf("\tchecking with dirent %s\n", dirent->d_name);
			if (strcmp(dirent->d_name, next_element) == 0) {
				/* get file or sub directory */
				errno = udf_lookup_node(udf_log_vol, &location, &cur_node);
				if (!errno) {
					udfclient_closedir(dirread);
					return udfclient_lookup_name(udf_log_vol, cur_node, restpath, dirent, new_node);
				};
			};
			udfclient_dirread_entry(dirread, dirent, &udf_log_vol, &location, NULL);
		};
		udfclient_closedir(dirread);
	};

	return ENOENT;
}


static void print_dir_entry(struct udf_icb_node *node, struct long_ad *dirnode_addr, struct dirent *dirent) {
	struct udf_icb_node *dirnode;
	uint64_t size;
	char *group, *owner, *name;
	int mode, this_mode;

	mode = 0744;
	if (dirent->d_type == DT_DIR) mode += 512;
	name = dirent->d_name;

	if (node) {
		errno = udf_lookup_node(node->udf_log_vol, dirnode_addr, &dirnode);
		size = udf_rw64(dirnode->dscrptr->fe.inf_len);
		group = owner = "unknown";
	} else {
		size = UDF_SECTOR_SIZE;
		group = owner = "udf";
	};

	if (mode & 512) printf("d"); else printf("-");
	mode = mode & 511;

	this_mode = (mode >> 6) & 7;
	printf("%c%c%c", "----rrrr"[this_mode & 4], "--www"[this_mode & 2], "-x"[this_mode & 1]);
	this_mode = (mode >> 3) & 7;
	printf("%c%c%c", "----rrrr"[this_mode & 4], "--www"[this_mode & 2], "-x"[this_mode & 1]);
	this_mode = mode & 7;
	printf("%c%c%c", "----rrrr"[this_mode & 4], "--www"[this_mode & 2], "-x"[this_mode & 1]);

	printf("    %s\t%s\t%10lu  %s\n", group, owner, (unsigned long) size, name);
}


void udfclient_ls(int args, char *arg1, char *arg2) {
	struct udfclient_dirread_stream *dirread;
	struct udf_icb_node *new_udf_node;
	struct long_ad dirnode_addr;
	struct dirent dirent;
	char *node_name;

	if (args > 2) {
		printf("Syntax: ls [file | dir]\n");
		return;
	};

	node_name = udfclient_realpath(curdir_name, arg1);

	errno = udfclient_lookup_name(NULL, NULL, node_name, &dirent, &new_udf_node);
	if (errno) {
		fprintf(stderr, "%s : ", arg1); perror("");
		free(node_name); 
		return;
	};

	if (dirent.d_type == DT_DIR) {
		printf("Directory listing of %s\n", dirent.d_name);
		errno = udfclient_opendir(new_udf_node, &dirread);
		udfclient_dirread_entry(dirread, &dirent, NULL, &dirnode_addr, NULL);
		while (!dirread->eof) {
			print_dir_entry(new_udf_node, &dirnode_addr, &dirent);

			udfclient_dirread_entry(dirread, &dirent, NULL, &dirnode_addr, NULL);
		};
		udfclient_closedir(dirread);
	} else {
		/* XXX dirnode_addr (long_ad) doesn't point to the fid of the selected file XXX */
		/* XXX FIX ME !!! XXX */
		//print_dir_entry(new_udf_node, &dirnode_addr, &dirent);
		printf("UDF: listing of single file not implemented yet (silly eh?)\n");
	};
	free(node_name);
}


void udfclient_pwd(int args) {
	char pwd[1024];

	if (args > 1) {
		printf("Syntax: pwd\n");
		return;
	};
	getcwd(pwd, 1024);
	printf("UDF working directory is     %s\n", curdir_name);
	printf("Current FS working directory %s\n", pwd);
}


void udfclient_cd(int args, char *arg1, char *arg2) {
	struct udf_icb_node *new_udf_node;
	struct dirent dirent;
	char *node_name, *new_curdir_name;

	if (args > 2) {
		printf("Syntax: cd [dir]\n");
		return;
	};
	new_curdir_name = udfclient_realpath(curdir_name, arg1);
	node_name = strdup(new_curdir_name);	/* working copy */

	errno = udfclient_lookup_name(NULL, NULL, node_name, &dirent, &new_udf_node);

	if (errno) {
		fprintf(stderr, "%s : ", arg1); perror("");
		free(node_name); 
		return;
	};

	if (dirent.d_type == DT_DIR) {
		free(curdir_name);
		curdir_name = new_curdir_name;
		curdir = new_udf_node;
		free(node_name);

		udfclient_pwd(0);
	} else {
		fprintf(stderr, "%s is not a directory\n", arg1);
		free(node_name);
		free(new_curdir_name);
	};
}


void udfclient_lcd(int args, char *arg1) {
	char pwd[1024];

	if (args > 2) {
		printf("Syntax: lcd [dir]\n");
		return;
	};
	if (strcmp(arg1, "")==0)  arg1 = getenv("HOME");
	if (strcmp(arg1, "~")==0) arg1 = getenv("HOME");

	if (chdir(arg1)) {
		fprintf(stderr, "While trying to go to %s :", arg1);
		perror("");
	};
	getcwd(pwd, 1024);
	printf("Changing local directory to %s\n", pwd);
}


void udfclient_getsubtree(char *name, struct udf_icb_node *udf_icb_node, int recursion) {
	struct udf_readdir_stream *readdir;
	struct udf_icb_node	  *sub_node;
	struct dirent	 entry;
	struct long_ad	 icb;
	uint8_t		*file_data;
	uint64_t	 file_data_got_len, file_data_total_len, data_pos;
	char		 fullpath[1024], buf[1024];	/* XXX */
	int		 fileh, len, udf_type, new_recurse;

	if (!udf_icb_node) return;
	if (udf_icb_node->dscrptr->fe.icbtag.file_type == UDF_ICB_FILETYPE_DIRECTORY && recursion) {
		errno = udf_opendir(udf_icb_node, 0, &readdir);
		if (!errno) {
			errno = udf_readdir_entry(readdir, &entry, &icb, &udf_type);
			while (!readdir->eof) {
				if (errno) perror("\treaddir_entry :");
				new_recurse = (strcmp(entry.d_name, ".") && strcmp(entry.d_name, ".."));
				errno = udf_lookup_node(udf_icb_node->udf_log_vol, &icb, &sub_node);
				if (errno) perror("\tlookup_node :");
				sprintf(fullpath, "%s/%s", name, entry.d_name);
				udfclient_getsubtree(fullpath, sub_node, new_recurse);

				errno = udf_readdir_entry(readdir, &entry, &icb, &udf_type);
			};
			udf_closedir(readdir);
		} else {
			perror("\topening directory :");
		};
		return;
	};

	/* terminal directory node? */
	if (udf_icb_node->dscrptr->fe.icbtag.file_type == UDF_ICB_FILETYPE_DIRECTORY) {
		strcpy(buf, name);

		len = strlen(buf);
		if (strcmp(buf+len-2, "/.")  == 0) buf[len-2] = 0;
		if (strcmp(buf+len-3, "/..") == 0) return;
		mkdir(buf, 0700);
		if (errno) perror("creating dir");
		return;
	};

	/* terminal file node */
	strcpy(buf, name);

	fileh = open(buf, O_WRONLY | O_CREAT, 0666);
	if (fileh>=0) {
		data_pos = file_data_got_len = 0;
		file_data_total_len = INT_MAX;
		do {
			errno = udf_read_file_part(udf_icb_node, data_pos, 256*1024, (void **) &file_data, &file_data_got_len, &file_data_total_len);
			if (errno) {
				perror("Got this error while retrieving file :");
				printf("\n\n\n");	/* XXX */
				break;
			};

			data_pos += file_data_got_len;
			write(fileh, file_data, file_data_got_len);
			free(file_data);

			if (strlen(name) < 50) {
				printf("\r%-50s ", name);
			} else {
				printf("\r...%-47s ", name+strlen(name)-47);
			};
			printf("%10d / %10d bytes ", (uint32_t) data_pos, (uint32_t) file_data_total_len);
			if (file_data_total_len) printf("(%3d%%) ", (int) (100.0*(float) data_pos / file_data_total_len));
			printf("successfully");
			fflush(stdout);
		} while ((data_pos < file_data_total_len));
		printf("\n");
		close(fileh);
	} else {
		printf("Help! can't open file %s for output\n", buf);
	};
}


void udfclient_get(int args, char *arg1, char *arg2) {
	struct udf_icb_node *udf_node;
	struct dirent dirent;
	char *node_name, *name;

	if (args != 2) {
		printf("Syntax: get (file | dir)\n");
		return;
	};

	node_name = udfclient_realpath(curdir_name, arg1);
	printf("Attempting to retrieve %s\n", node_name);

	errno = udfclient_lookup_name(NULL, NULL, node_name, &dirent, &udf_node);

	if (errno) {
		fprintf(stderr, "%s : ", arg1); perror("");
		free(node_name); 
		return;
	};

	name = strdup(dirent.d_name);
	udfclient_getsubtree(name, udf_node, 1);
	free(node_name);
};


void udfclient_interact(void) {
	int  args;
	char cmd[1024], arg1[1024], arg2[1024], arg3[1024];
	char line[4096];

	udfclient_pwd(0);
	while (1) {
		printf("UDF> ");
		clearerr(stdin);

		*line = 0;
		fgets(line, 4096, stdin);

		*cmd = *arg1 = *arg2 = *arg3 = 0;
		args = sscanf(line, "%s %s %s %s", cmd, arg1, arg2, arg3);
		if (args) {
			if (strcmp(cmd, "")==0) continue;

			if (strcmp(cmd, "ls")==0) {
				udfclient_ls(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "cd")==0) {
				udfclient_cd(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "lcd")==0) {
				udfclient_lcd(args, arg1);
				continue;
			};
			if (strcmp(cmd, "pwd")==0) {
				udfclient_pwd(args);
				continue;
			};
			if (strcmp(cmd, "get")==0) {
				udfclient_get(args, arg1, arg2);
				continue;
			};
			if (strcmp(cmd, "help")==0) {
				printf("Selected commands available:\n"
						"ls  [file | dir]\t\tlists a directory\n"
						"cd  [dir]\t\tchange current UDF directory\n"
						"lcd [dir]\t\tchange current directory\n"
						"pwd\t\t\tdisplay current directories\n"
						"get (file | dir)\tretrieve a file or a directory\n"
						"quit\t\t\texits program\n"
				      );
				continue;
			};
			if (strcmp(cmd, "quit")==0) {
				return;
			};
			printf("\nUnknown command %s\n", cmd);
		};
	};
}


int usage(char *program) {
	fprintf(stderr, "Usage: %s [options] devicename [devicename]*)\n", program);
	fprintf(stderr, "-u level	UDF system verbose level\n"
			"-s 		verbose SCSI command errors\n"
			"-r range	use only selected sessions like -3,5,7 or 6-\n"
	);
	return 1;
}


extern char	*optarg;
extern int	 optind;
extern int	 optreset;


int main(int argc, char *argv[]) {
	struct udf_discinfo *disc, *next_disc;
	char   *progname, *range, flag;

	progname = argv[0];
	if (argc == 1) {
		return usage(progname);
	};

	udf_verbose	= UDF_VERBLEV_NONE;
	scsilib_verbose	= 0;
	range		= NULL;
	while ((flag = getopt(argc, argv, "u:sr:")) != -1) {
		switch (flag) {
			case 'u' :
				udf_verbose = atoi(optarg);
				break;
			case 's' :
				scsilib_verbose = 1;
				break;
			case 'r' :
				range = strdup(optarg);
				if (udf_check_session_range(range)) {
					fprintf(stderr, "Invalid range %s\n", range);
					return usage(progname);
				};
				break;
			default  :
				return usage(progname);
		};
	};
	argv += optind;
	argc -= optind;

	if (argc == 0) return usage(progname);

	/* all other arguments are devices */
	udf_discs_list = NULL;
	while (argc) {
		printf("Opening device %s\n", *argv);
		errno = udf_open_and_read_disc(*argv, range, &disc);
		if (errno) {
			perror("Can't open my device; bailing out ");
			exit(1);
		};

		disc->next = udf_discs_list;
		udf_discs_list = disc;
		argv++; argc--;
		if (udf_verbose) printf("\n\n");
	};

	printf("\n");
	printf("Resulting list of alive sets :\n\n");
	udf_dump_alive_sets();

	/* interactive part */
	curdir = NULL;
	curdir_name = strdup("/");
	udfclient_interact();

	/* close part */
	printf("Closing discs\n");
	disc = udf_discs_list;
	while (disc) {
		next_disc = disc->next;
		udf_close_disc(disc);
		disc = next_disc;
	};

	return 0;
}

